from optinist.wrappers.caiman_wrapper import caiman_wrapper_dict
from optinist.wrappers.lccd_wrapper import lccd_wrapper_dict

# from optinist.wrappers.dummy_wrapper import dummy_wrapper_dict
from optinist.wrappers.optinist_wrapper import optinist_wrapper_dict
from optinist.wrappers.suite2p_wrapper import suite2p_wrapper_dict

wrapper_dict = {}
wrapper_dict.update(**caiman_wrapper_dict)
wrapper_dict.update(**suite2p_wrapper_dict)
# wrapper_dict.update(**dummy_wrapper_dict)
wrapper_dict.update(**optinist_wrapper_dict)
wrapper_dict.update(**lccd_wrapper_dict)



import h5py

from optinist.api.dataclass.dataclass import CsvData, ImageData, TimeSeriesData
from optinist.api.nwb.nwb import NWBDATASET
from optinist.api.snakemake.smk import Rule
from optinist.routers.model import FILETYPE


class FileWriter:
    @classmethod
    def csv(cls, rule_config: Rule, nodeType):
        info = {
            rule_config.return_arg: CsvData(rule_config.input, rule_config.params, "")
        }
        nwbfile = rule_config.nwbfile

        if nodeType == FILETYPE.CSV:
            if NWBDATASET.TIMESERIES not in nwbfile:
                nwbfile[NWBDATASET.TIMESERIES] = {}
            nwbfile[NWBDATASET.TIMESERIES][rule_config.return_arg] = info[
                rule_config.return_arg
            ]
        elif nodeType == FILETYPE.BEHAVIOR:
            if NWBDATASET.BEHAVIOR not in nwbfile:
                nwbfile[NWBDATASET.BEHAVIOR] = {}
            nwbfile[NWBDATASET.BEHAVIOR][rule_config.return_arg] = info[
                rule_config.return_arg
            ]
        else:
            assert False, "NodeType doesn't exsits"

        nwbfile.pop("image_series", None)
        info["nwbfile"] = {"input": nwbfile}
        return info

    @classmethod
    def image(cls, rule_config: Rule):
        info = {rule_config.return_arg: ImageData(rule_config.input, "")}
        nwbfile = rule_config.nwbfile
        nwbfile["image_series"]["external_file"] = info[rule_config.return_arg]
        info["nwbfile"] = {"input": nwbfile}
        return info

    @classmethod
    def hdf5(cls, rule_config: Rule):
        nwbfile = rule_config.nwbfile

        with h5py.File(rule_config.input, "r") as f:
            data = f[rule_config.hdf5Path][:]

        if data.ndim == 3:
            info = {rule_config.return_arg: ImageData(data, "")}
            nwbfile["image_series"]["external_file"] = info[rule_config.return_arg]
            info["nwbfile"] = {}
            info["nwbfile"][FILETYPE.IMAGE] = nwbfile
        elif data.ndim == 2:
            info = {rule_config.return_arg: TimeSeriesData(data, "")}

            if NWBDATASET.TIMESERIES not in nwbfile:
                nwbfile[NWBDATASET.TIMESERIES] = {}

            nwbfile[NWBDATASET.TIMESERIES][rule_config.return_arg] = info[
                rule_config.return_arg
            ]
            nwbfile.pop("image_series", None)
            info["nwbfile"] = {"input": nwbfile}

        return info



import numpy as np
import scipy
import scipy.signal
import skimage
import skimage.exposure
import skimage.filters
import skimage.measure

from optinist.wrappers.lccd_wrapper.lccd_python import oval_filter
from optinist.wrappers.lccd_wrapper.lccd_python.utils import matlab_conv2


def matlab_style_gauss2D(shape=(3, 3), sigma=0.5):
    """
    2D gaussian mask - should give the same result as MATLAB's
    fspecial('gaussian',[shape],[sigma])
    From
    https://stackoverflow.com/questions/17190649/how-to-obtain-a-gaussian-filter-in-python
    """
    m, n = [(ss - 1.0) / 2.0 for ss in shape]
    y, x = np.ogrid[-m : m + 1, -n : n + 1]
    h = np.exp(-(x * x + y * y) / (2.0 * sigma * sigma))
    h[h < np.finfo(h.dtype).eps * h.max()] = 0
    sumh = h.sum()
    if sumh != 0:
        h /= sumh
    return h


def im2bw(nd_array_img, level):
    result = np.zeros_like(nd_array_img)
    result[nd_array_img > level] = 1
    return result


def filter_closed_regions_by_area(mat, min_area, max_area):
    mat = np.copy(mat)
    num_regions = np.max(mat)
    for i in range(1, 1 + num_regions):
        area = np.sum(mat == i)
        if not (min_area < area and area <= max_area):
            mat[mat == i] = 0
    return mat


class BlobDetector:
    def __init__(
        self,
        filtersize1,
        filtersize2,
        sigma,
        fsize,
        min_area,
        max_area,
        matlab_conv=False,
        debug=False,
        sparse=True,
        **kwargs
    ):
        self.min_area = min_area
        self.max_area = max_area
        self.mavefil1 = np.ones((1, filtersize1)) / filtersize1
        self.mavefil2 = np.ones((1, filtersize2)) / filtersize2
        self.sp_fil = matlab_style_gauss2D(shape=(fsize, fsize), sigma=sigma) - np.ones(
            (fsize, fsize)
        ) / (
            fsize**2
        )  # average filter
        self.conv2 = matlab_conv2 if matlab_conv else scipy.signal.convolve2d
        self.debug = debug
        self.sparse = sparse

    def apply(self, tv):
        X, Y, T = tv.shape
        tv = tv.reshape(-1, T)
        smooth_tv = np.maximum(
            self.conv2(tv, self.mavefil2, mode="same")
            - self.conv2(tv, self.mavefil1, mode="same"),
            0,
        )
        max_lumi_map = np.max(smooth_tv, axis=1).reshape(X, Y)
        lumi_map = np.maximum(
            self.conv2(max_lumi_map, self.sp_fil, mode="same"),
            # Here, the article says the mode is 'valid' (Algorithm2, line 10).
            # But in order that ROI have the dimesion (0, 1)^{XY x N_r} (line 21),
            # mode should be 'same'.
            0,
        )
        cont_enhanced_map = skimage.exposure.equalize_adapthist(lumi_map)
        threshold = skimage.filters.threshold_otsu(cont_enhanced_map)
        bwm = im2bw(cont_enhanced_map, threshold)
        label_mat = skimage.measure.label(bwm, connectivity=2)
        label_mat2 = filter_closed_regions_by_area(
            label_mat, min_area=self.min_area, max_area=self.max_area
        )
        ROI = oval_filter.oval_filter(label_mat2, sparse=self.sparse)
        if not self.debug:
            return ROI
        return {
            "smooth_tv": smooth_tv,
            "max_lumi_map": max_lumi_map,
            "lumi_map": lumi_map,
            "cont_enhanced_map": cont_enhanced_map,
            "threshold": threshold,
            "bwm": bwm,
            "label_mat": label_mat,
            "label_mat2": label_mat2,
            "ROI": ROI,
            "lumi_map_shape": lumi_map.shape,
        }



from optinist.api.dataclass.dataclass import FluoData, HeatMapData, IscellData
from optinist.api.nwb.nwb import NWBDATASET


def correlation(
    neural_data: FluoData,
    output_dir: str,
    iscell: IscellData = None,
    params: dict = None,
) -> dict():
    import numpy as np

    neural_data = neural_data.data

    # data shold be time x component matrix
    if params["transpose"]:
        X = neural_data.transpose()
    else:
        X = neural_data

    if iscell is not None:
        iscell = iscell.data
        ind = np.where(iscell > 0)[0]
        X = X[ind, :]

    num_cell = X.shape[0]

    # calculate correlation
    corr = np.corrcoef(X)
    for i in range(num_cell):
        corr[i, i] = np.nan

    # NWB追加
    nwbfile = {}
    nwbfile[NWBDATASET.POSTPROCESS] = {
        "corr": corr,
    }

    info = {
        "corr": HeatMapData(corr, file_name="corr"),
        "nwbfile": nwbfile,
    }

    return info



# import argparse
# import joblib
# import json

import numpy as np

from optinist.wrappers.lccd_wrapper.lccd_python.blob_detector import BlobDetector
from optinist.wrappers.lccd_wrapper.lccd_python.roi_integration import RoiIntegration


class LCCD:
    def __init__(self, config):
        self.blob_detector = BlobDetector(**config["blob_detector"])
        self.roi_integration = RoiIntegration(**config["roi_integration"])
        self.frame_divider = config["lccd"]["frame_divider"]

    def apply(self, v):
        # lazy loading of v will make this method memory efficient.
        # https://qiita.com/Hanjin_Liu/items/7a01c1c481161fe20a3e
        roi = None
        X, Y, T = v.shape
        for m in range(0, T // self.frame_divider - 1):
            tm = v[:, :, m * self.frame_divider : (m + 1) * self.frame_divider]
            roi_tm = self.blob_detector.apply(tm)
            roi = self.roi_integration.apply(roi_tm, roi)
        if self.roi_integration.sparse and roi is not None:
            roi = np.array(roi.todense())
        return roi


# def main(args):
#     # args = parser.parse_args()
#     with open("config/config.json", "r") as f:
#         config = json.load(f)
#     lccd = LCCD(config)

#     v = np.load("profiles/src.npy")
#     assert len(v.shape) == 3, (
#         "input array should have"
#         "dimensions (width, height, time)"
#     )
#     roi = lccd.apply(v)
#     joblib.dump({"config": config, "roi": roi}, "out.dump")


# if __name__ == "__main__":
#     parser = argparse.ArgumentParser()
#     parser.add_argument(
#         "--conf", default="config/config.json",
#         help="lccd config file"
#     )
#     parser.add_argument("--src", help="path to input file.")
#     parser.add_argument("--dst", help="path to output file.")
#     args = parser.parse_args()
#     main(args)



from optinist.api.dataclass.dataclass import ImageData, Suite2pData


def suite2p_registration(
    ops: Suite2pData, output_dir: str, params: dict = None
) -> dict(ops=Suite2pData):
    from suite2p import default_ops, registration

    ops = ops.data
    refImg = ops["meanImg"]
    print("start suite2_registration")

    # REGISTRATION
    if len(refImg.shape) == 3:
        refImg = refImg[0]

    ops = {**default_ops(), **ops, **params}

    # register binary
    ops = registration.register_binary(ops, refImg=refImg)

    # compute metrics for registration
    if ops.get("do_regmetrics", True) and ops["nframes"] >= 1500:
        ops = registration.get_pc_metrics(ops)

    info = {
        "refImg": ImageData(ops["refImg"], output_dir=output_dir, file_name="refImg"),
        "meanImgE": ImageData(
            ops["meanImgE"], output_dir=output_dir, file_name="meanImgE"
        ),
        "ops": Suite2pData(ops, file_name="ops"),
    }

    return info



from optinist.wrappers.caiman_wrapper.cnmf import caiman_cnmf
from optinist.wrappers.caiman_wrapper.motion_correction import caiman_mc

caiman_wrapper_dict = {
    "caiman": {
        "caiman_mc": {
            "function": caiman_mc,
            "conda_name": "caiman",
            "conda_yaml": "caiman_env.yaml",
        },
        "caiman_cnmf": {
            "function": caiman_cnmf,
            "conda_name": "caiman",
            "conda_yaml": "caiman_env.yaml",
        },
    }
}



from optinist.api.experiment.experiment import ExptConfig


class ExptConfigBuilder:
    def __init__(self):
        self._timestamp = None
        self._name = None
        self._unique_id = None
        self._hasNWB = False
        self._function = {}
        self._nodeDict = None
        self._edgeDict = None

    def set_config(self, config: ExptConfig) -> "ExptConfigBuilder":
        self._timestamp = config.timestamp
        self._name = config.name
        self._unique_id = config.unique_id
        self._hasNWB = config.hasNWB
        self._function = config.function
        self._nodeDict = config.nodeDict
        self._edgeDict = config.edgeDict
        return self

    def set_timestamp(self, timestamp) -> "ExptConfigBuilder":
        self._timestamp = timestamp
        return self

    def set_name(self, name) -> "ExptConfigBuilder":
        self._name = name
        return self

    def set_unique_id(self, unique_id) -> "ExptConfigBuilder":
        self._unique_id = unique_id
        return self

    def set_hasNWB(self, hasNWB) -> "ExptConfigBuilder":
        self._hasNWB = hasNWB
        return self

    def set_function(self, function) -> "ExptConfigBuilder":
        self._function = function
        return self

    def set_nodeDict(self, nodeDict) -> "ExptConfigBuilder":
        self._nodeDict = nodeDict
        return self

    def set_edgeDict(self, edgeDict) -> "ExptConfigBuilder":
        self._edgeDict = edgeDict
        return self

    def build(self) -> ExptConfig:
        return ExptConfig(
            timestamp=self._timestamp,
            name=self._name,
            unique_id=self._unique_id,
            hasNWB=self._hasNWB,
            function=self._function,
            nodeDict=self._nodeDict,
            edgeDict=self._edgeDict,
        )



from dataclasses import dataclass
from typing import Dict, List, Union

from pydantic import BaseModel

from optinist.api.snakemake.smk import ForceRun


@dataclass
class NodeType:
    IMAGE: str = "ImageFileNode"
    CSV: str = "CsvFileNode"
    FLUO: str = "FluoFileNode"
    BEHAVIOR: str = "BehaviorFileNode"
    HDF5: str = "HDF5FileNode"
    ALGO: str = "AlgorithmNode"


@dataclass
class OutputType:
    IMAGE: str = "images"
    TIMESERIES: str = "timeseries"
    HEATMAP: str = "heatmap"
    ROI: str = "roi"
    SCATTER: str = "scatter"
    BAR: str = "bar"
    HTML: str = "html"


class NodeItem(BaseModel):
    pendingNodeIdList: list = []


@dataclass
class OutputPath:
    path: str
    type: str
    max_index: int = None


@dataclass
class Message:
    status: str
    message: str
    outputPaths: Dict[str, OutputPath] = None


@dataclass
class NodeData:
    label: str
    param: dict
    path: Union[str, List]
    type: str
    fileType: str = None
    hdf5Path: str = None


@dataclass
class NodePosition:
    x: int
    y: int


@dataclass
class Style:
    border: str = None
    height: int = None
    padding: int = None
    width: int = None
    borderRadius: int = None


@dataclass
class Node:
    id: str
    type: str
    data: NodeData
    position: NodePosition
    style: Style


@dataclass
class Edge:
    id: str
    type: str
    animated: bool
    source: str
    sourceHandle: str
    target: str
    targetHandle: str
    style: Style


class RunItem(BaseModel):
    name: str = None
    nodeDict: dict = {}
    edgeDict: dict = {}
    snakemakeParam: dict = {}
    nwbParam: dict = {}
    forceRunList: List[ForceRun]



import os
from glob import glob
from typing import Optional

import pandas as pd
from fastapi import APIRouter

from optinist.api.dir_path import DIRPATH
from optinist.api.edit_ROI import ACTION, EditROI
from optinist.api.utils.filepath_creater import create_directory, join_filepath
from optinist.api.utils.json_writer import JsonWriter, save_tiff2json
from optinist.routers.const import ACCEPT_TIFF_EXT
from optinist.routers.fileIO.file_reader import JsonReader, Reader
from optinist.routers.model import (
    EditRoiSuccess,
    JsonTimeSeriesData,
    OutputData,
    RoiList,
    RoiPos,
)

router = APIRouter()


@router.get(
    "/outputs/inittimedata/{dirpath:path}",
    response_model=JsonTimeSeriesData,
    tags=["outputs"],
)
async def get_inittimedata(dirpath: str):
    file_numbers = sorted(
        [
            os.path.splitext(os.path.basename(x))[0]
            for x in glob(join_filepath([dirpath, "*.json"]))
        ]
    )

    index = file_numbers[0]
    str_index = str(index)

    json_data = JsonReader.read_as_timeseries(
        join_filepath([dirpath, f"{str(index)}.json"])
    )

    return_data = JsonTimeSeriesData(
        xrange=[],
        data={},
        std={},
    )

    data = {
        str(i): {json_data.xrange[0]: json_data.data[json_data.xrange[0]]}
        for i in file_numbers
    }

    if json_data.std is not None:
        std = {
            str(i): {json_data.xrange[0]: json_data.data[json_data.xrange[0]]}
            for i in file_numbers
        }

    return_data = JsonTimeSeriesData(
        xrange=json_data.xrange,
        data=data,
        std=std if json_data.std is not None else {},
    )

    return_data.data[str_index] = json_data.data
    if json_data.std is not None:
        return_data.std[str_index] = json_data.std

    return return_data


@router.get(
    "/outputs/timedata/{dirpath:path}",
    response_model=JsonTimeSeriesData,
    tags=["outputs"],
)
async def get_timedata(dirpath: str, index: int):
    json_data = JsonReader.read_as_timeseries(
        join_filepath([dirpath, f"{str(index)}.json"])
    )

    return_data = JsonTimeSeriesData(
        xrange=[],
        data={},
        std={},
    )

    str_index = str(index)
    return_data.data[str_index] = json_data.data
    if json_data.std is not None:
        return_data.std[str_index] = json_data.std

    return return_data


@router.get(
    "/outputs/alltimedata/{dirpath:path}",
    response_model=JsonTimeSeriesData,
    tags=["outputs"],
)
async def get_alltimedata(dirpath: str):
    return_data = JsonTimeSeriesData(
        xrange=[],
        data={},
        std={},
    )

    for i, path in enumerate(glob(join_filepath([dirpath, "*.json"]))):
        str_idx = str(os.path.splitext(os.path.basename(path))[0])
        json_data = JsonReader.read_as_timeseries(path)
        if i == 0:
            return_data.xrange = json_data.xrange

        return_data.data[str_idx] = json_data.data
        if json_data.std is not None:
            return_data.std[str_idx] = json_data.std

    return return_data


@router.get(
    "/outputs/data/{filepath:path}", response_model=OutputData, tags=["outputs"]
)
async def get_file(filepath: str):
    return JsonReader.read_as_output(filepath)


@router.get(
    "/outputs/html/{filepath:path}", response_model=OutputData, tags=["outputs"]
)
async def get_html(filepath: str):
    return Reader.read_as_output(filepath)


@router.get(
    "/outputs/image/{filepath:path}", response_model=OutputData, tags=["outputs"]
)
async def get_image(
    filepath: str, start_index: Optional[int] = 0, end_index: Optional[int] = 1
):
    filename, ext = os.path.splitext(os.path.basename(filepath))
    if ext in ACCEPT_TIFF_EXT:
        filepath = join_filepath([DIRPATH.INPUT_DIR, filepath])
        save_dirpath = join_filepath(
            [
                os.path.dirname(filepath),
                filename,
            ]
        )
        json_filepath = join_filepath(
            [save_dirpath, f"{filename}_{str(start_index)}_{str(end_index)}.json"]
        )
        if not os.path.exists(json_filepath):
            save_tiff2json(filepath, save_dirpath, start_index, end_index)
    else:
        json_filepath = filepath

    return JsonReader.read_as_output(json_filepath)


@router.post(
    "/outputs/image/{filepath:path}/add_roi",
    response_model=EditRoiSuccess,
    tags=["outputs"],
)
async def add_roi(filepath: str, pos: RoiPos):
    EditROI(action=ACTION.ADD, filepath=filepath, params=pos.dict()).excute()
    return EditRoiSuccess(max_index=0)


@router.post(
    "/outputs/image/{filepath:path}/merge_roi",
    response_model=EditRoiSuccess,
    tags=["outputs"],
)
async def merge_roi(filepath: str, roi_list: RoiList):
    EditROI(action=ACTION.MERGE, filepath=filepath, params=roi_list.dict()).excute()
    return EditRoiSuccess(max_index=0)


@router.post(
    "/outputs/image/{filepath:path}/delete_roi",
    response_model=EditRoiSuccess,
    tags=["outputs"],
)
async def delete_roi(filepath: str, roi_list: RoiList):
    EditROI(action=ACTION.DELETE, filepath=filepath, params=roi_list.dict()).excute()
    return EditRoiSuccess(max_index=0)


@router.get("/outputs/csv/{filepath:path}", response_model=OutputData, tags=["outputs"])
async def get_csv(filepath: str):
    filepath = join_filepath([DIRPATH.INPUT_DIR, filepath])

    filename, _ = os.path.splitext(os.path.basename(filepath))
    save_dirpath = join_filepath([os.path.dirname(filepath), filename])
    create_directory(save_dirpath)
    json_filepath = join_filepath([save_dirpath, f"{filename}.json"])

    JsonWriter.write_as_split(json_filepath, pd.read_csv(filepath, header=None))
    return JsonReader.read_as_output(json_filepath)



class AlgorithmException(Exception):
    """アルゴリズムの実行時に起きる例外の基底クラス"""

    def __init__(self, message: str):
        super().__init__()
        self._message = message

    def get_message(self) -> str:
        return self._message


class ArgsMissingException(AlgorithmException):
    pass


class ArgsTypeException(AlgorithmException):
    pass



from optinist.wrappers.optinist_wrapper.neural_population_analysis.correlation import (
    correlation,
)
from optinist.wrappers.optinist_wrapper.neural_population_analysis.cross_correlation import (  # noqa: E501
    cross_correlation,
)
from optinist.wrappers.optinist_wrapper.neural_population_analysis.granger import (
    Granger,
)

neural_population_analysis_wrapper_dict = {
    "correlation": {
        "function": correlation,
    },
    "cross_correlation": {
        "function": cross_correlation,
        "conda_yaml": "optinist_env.yaml",
    },
    "granger": {
        "function": Granger,
        "conda_yaml": "optinist_env.yaml",
    },
}



from optinist.api.dataclass.dataclass import FluoData, IscellData, TimeSeriesData
from optinist.api.nwb.nwb import NWBDATASET


def cross_correlation(
    neural_data: FluoData,
    output_dir: str,
    iscell: IscellData = None,
    params: dict = None,
) -> dict():
    import itertools

    import numpy as np
    import scipy.signal as ss
    import scipy.stats as stats
    from tqdm import tqdm

    neural_data = neural_data.data

    # data shold be time x component matrix
    if params["transpose"]:
        X = neural_data.transpose()
    else:
        X = neural_data

    if iscell is not None:
        iscell = iscell.data
        ind = np.where(iscell > 0)[0]
        X = X[ind, :]

    # calculate cross correlation
    num_cell = X.shape[0]
    data_len = X.shape[1]
    shuffle_num = params["shuffle_sample_number"]

    lags = ss.correlation_lags(data_len, data_len, mode="same")
    ind = np.intersect1d(
        np.where(-params["lags"] <= lags), np.where(lags <= params["lags"])
    )
    x = lags[ind]

    mat = np.zeros([num_cell, num_cell, len(x)])
    s_confint = np.zeros([num_cell, num_cell, len(x), 2])
    s_mean = np.zeros([num_cell, num_cell, len(x)])

    for i in tqdm(range(num_cell)):
        # for i in tqdm(range(3)):
        for j in tqdm(range(num_cell)):
            ccvals = ss.correlate(
                X[i, :], X[j, :], method=params["method"], mode="same"
            )
            mat[i, j, :] = ccvals[ind]

            # baseline
            tp_s_mat = np.zeros([len(x), shuffle_num])
            for k in range(shuffle_num):
                tp = X[j, :]
                np.random.shuffle(tp)
                ccvals = ss.correlate(X[i, :], tp, method=params["method"], mode="same")
                tp_s_mat[:, k] = ccvals[ind]

            b_mean = np.mean(tp_s_mat, axis=1)
            b_sem = stats.sem(tp_s_mat, axis=1)
            b_df = shuffle_num - 1

            interval = np.zeros([len(x), 2])
            for k in range(len(x)):
                interval[k, :] = stats.t.interval(
                    params["shuffle_confidence_interval"], b_df, loc=0, scale=b_sem[k]
                )

            s_confint[i, j, :, 0] = interval[:, 0]
            s_confint[i, j, :, 1] = interval[:, 1]
            s_mean[i, j, :] = b_mean

    # NWB追加
    nwbfile = {}
    nwbfile[NWBDATASET.POSTPROCESS] = {
        "mat": mat,
        "baseline": s_mean,
        "base_confint": s_confint,
    }

    info = {
        # 'cross_correlation': ScatterData(mat),
        "nwbfile": nwbfile
    }

    # output structures
    cb = list(itertools.combinations(range(num_cell), 2))

    for i in range(len(cb)):
        arr1 = np.stack([x, mat[cb[i][0], cb[i][1], :]], axis=1)
        arr2 = np.stack(
            [
                x,
                s_mean[cb[i][0], cb[i][1], :],
                s_confint[cb[i][0], cb[i][1], :, 0],
                s_confint[cb[i][0], cb[i][1], :, 1],
            ],
            axis=1,
        )

    name = f"{str(cb[i][0])}-{str(cb[i][1])}"
    info[name] = TimeSeriesData(arr1.T, file_name=name)
    name = f"shuffle {str(cb[i][0])}-{str(cb[i][1])}"
    info[name] = TimeSeriesData(arr2.T, file_name=name)

    return info



from optinist.api.dataclass.dataclass import (
    FluoData,
    HeatMapData,
    IscellData,
    ScatterData,
)
from optinist.api.nwb.nwb import NWBDATASET
from optinist.wrappers.optinist_wrapper.utils import standard_norm


def Granger(
    neural_data: FluoData,
    output_dir: str,
    iscell: IscellData = None,
    params: dict = None,
) -> dict():
    # modules specific to function
    # from sklearn.preprocessing import StandardScaler
    import itertools

    import numpy as np
    from statsmodels.tsa.stattools import adfuller, coint, grangercausalitytests
    from tqdm import tqdm

    neural_data = neural_data.data

    # data shold be time x component matrix
    if params["transpose"]:
        X = neural_data.transpose()
    else:
        X = neural_data

    if iscell is not None:
        iscell = iscell.data
        ind = np.where(iscell > 0)[0]
        X = X[:, ind]

    num_cell = X.shape[1]
    comb = list(itertools.permutations(range(num_cell), 2))  # combinations with dup
    num_comb = len(comb)

    # preprocessing
    tX = standard_norm(X, params["standard_mean"], params["standard_std"])

    # calculate dickey-fuller test
    # augmented dickey-fuller test
    # - if p val is large
    #   -> it cannot reject  there is a unit root
    # - small p-val means OK
    #   -> means this is not unit root process that it can apply Causality test

    adf = {
        "adf_teststat": np.zeros([num_cell], dtype="float64"),
        "adf_pvalue": np.zeros([num_cell], dtype="float64"),
        "adf_usedlag": np.zeros([num_cell], dtype="int"),
        "adf_nobs": np.zeros([num_cell], dtype="int"),
        "adf_critical_values": np.zeros([num_cell, 3], dtype="float64"),
        "adf_icbest": np.zeros([num_cell], dtype="float64"),
    }

    if params["use_adfuller_test"]:
        print("adfuller test ")

        for i in tqdm(range(num_cell)):
            tp = adfuller(tX[:, i], **params["adfuller"])

            adf["adf_teststat"][i] = tp[0]
            adf["adf_pvalue"][i] = tp[1]
            adf["adf_usedlag"][i] = tp[2]
            adf["adf_nobs"][i] = tp[3]
            adf["adf_critical_values"][i, :] = np.array(
                [tp[4]["1%"], tp[4]["5%"], tp[4]["10%"]]
            )
            adf["adf_icbest"][i] = tp[5]

    #  test for cointegration
    # augmented engle-granger two-step test
    # Test for no-cointegration of a univariate equation
    # if p val is small, the relation is cointegration
    # -> check this if ADF pval is large
    cit = {
        "cit_count_t": np.zeros([num_comb], dtype="float64"),
        "cit_pvalue": np.zeros([num_comb], dtype="int"),
        "cit_crit_value": np.zeros([num_comb, 3], dtype="float64"),
    }

    if params["use_coint_test"]:
        print("cointegration test ")

        for i in tqdm(range(num_comb)):
            tp = coint(X[:, comb[i][0]], X[:, comb[i][1]], **params["coint"])
            if not np.isnan(tp[0]):
                cit["cit_count_t"][i] = tp[0]
            if not np.isnan(tp[1]):
                cit["cit_pvalue"][i] = tp[1]
            cit["cit_crit_value"][i, :] = tp[2]

    #  Granger causality
    print("granger test ")

    if hasattr(params["Granger_maxlag"], "__iter__"):
        num_lag = len(params["Granger_maxlag"])
    else:
        num_lag = params["Granger_maxlag"]

    GC = {
        "gc_combinations": comb,
        "gc_ssr_ftest": np.zeros([num_comb, num_lag, 4], dtype="float64"),
        "gc_ssr_chi2test": np.zeros([num_comb, num_lag, 3], dtype="float64"),
        "gc_lrtest": np.zeros([num_comb, num_lag, 3], dtype="float64"),
        "gc_params_ftest": np.zeros([num_comb, num_lag, 4], dtype="float64"),
        "gc_OLS_restricted": [[0] * num_lag for i in range(num_comb)],
        "gc_OLS_unrestricted": [[0] * num_lag for i in range(num_comb)],
        "gc_OLS_restriction_matrix": [[0] * num_lag for i in range(num_comb)],
        "Granger_fval_mat": [np.zeros([num_cell, num_cell]) for i in range(num_lag)],
    }

    for i in tqdm(range(len(comb))):
        # The Null hypothesis for grangercausalitytests is
        # that the time series in the second column1,
        # does NOT Granger cause the time series in the first column0
        # column 1 -> colum 0
        tp = grangercausalitytests(
            tX[:, [comb[i][0], comb[i][1]]],
            params["Granger_maxlag"],
            verbose=False,
            addconst=params["Granger_addconst"],
        )

        for j in range(len(tp)):  # number of lag
            GC["gc_ssr_ftest"][i, j, :] = tp[j + 1][0]["ssr_ftest"][
                0:4
            ]  # ssr based F test (F, pval, df_denom, df_num)
            GC["gc_ssr_chi2test"][i, j, :] = tp[j + 1][0]["ssr_chi2test"][
                0:3
            ]  # ssr based chi2test (chi2, pval, df)
            GC["gc_lrtest"][i, j, :] = tp[j + 1][0]["lrtest"][
                0:3
            ]  # likelihood ratio test (chi2, pval, df)
            GC["gc_params_ftest"][i, j, :] = tp[j + 1][0]["params_ftest"][
                0:4
            ]  # parameter F test (F, pval, df_denom, df_num)
            GC["gc_OLS_restricted"][i][j] = tp[j + 1][1][0]
            GC["gc_OLS_unrestricted"][i][j] = tp[j + 1][1][1]
            GC["gc_OLS_restriction_matrix"][i][j] = tp[j + 1][1][2]

            GC["Granger_fval_mat"][j][comb[i][0], comb[i][1]] = tp[j + 1][0][
                "ssr_ftest"
            ][0]

    GC["Granger_fval_mat"] = np.array(GC["Granger_fval_mat"])

    # main results for plot
    info = {}
    info["Granger_fval_mat_heatmap"] = HeatMapData(
        GC["Granger_fval_mat"][0], file_name="gfm_heatmap"
    )
    info["Granger_fval_mat_scatter"] = ScatterData(
        GC["Granger_fval_mat"][0], file_name="gfm"
    )

    # NWB追加
    nwbfile = {}
    nwbfile[NWBDATASET.POSTPROCESS] = {
        "Granger_fval_mat": GC["Granger_fval_mat"][0],
        "gc_combinations": GC["gc_combinations"],
        "gc_ssr_ftest": GC["gc_ssr_ftest"],
        "gc_ssr_chi2test": GC["gc_ssr_chi2test"],
        "gc_lrtest": GC["gc_lrtest"],
        "gc_params_ftest": GC["gc_params_ftest"],
        "cit_pvalue": cit["cit_pvalue"],
        "adf_pvalue": adf["adf_pvalue"],
    }

    info["nwbfile"] = nwbfile

    return info



from optinist.api.edit_ROI.wrappers.caiman_edit_roi import (
    excute_delete_roi as caiman_delete,
)
from optinist.api.edit_ROI.wrappers.caiman_edit_roi import execute_add_ROI as caiman_add
from optinist.api.edit_ROI.wrappers.caiman_edit_roi import (
    execute_merge_roi as caiman_merge,
)
from optinist.api.edit_ROI.wrappers.lccd_edit_roi import (
    excute_delete_roi as lccd_delete,
)
from optinist.api.edit_ROI.wrappers.lccd_edit_roi import execute_add_ROI as lccd_add
from optinist.api.edit_ROI.wrappers.lccd_edit_roi import execute_merge_roi as lccd_merge
from optinist.api.edit_ROI.wrappers.suite2p_edit_roi import (
    excute_delete_roi as suite2p_delete,
)
from optinist.api.edit_ROI.wrappers.suite2p_edit_roi import (
    execute_add_ROI as suite2p_add,
)
from optinist.api.edit_ROI.wrappers.suite2p_edit_roi import (
    execute_merge_roi as suite2p_merge,
)

edit_roi_wrapper_dict = {
    "suite2p": {
        "conda_yaml": "suite2p_env.yaml",
        "function": {
            "add": suite2p_add,
            "merge": suite2p_merge,
            "delete": suite2p_delete,
        },
    },
    "lccd": {
        "conda_yaml": None,
        "function": {
            "add": lccd_add,
            "merge": lccd_merge,
            "delete": lccd_delete,
        },
    },
    "caiman": {
        "conda_yaml": None,
        "function": {
            "add": caiman_add,
            "merge": caiman_merge,
            "delete": caiman_delete,
        },
    },
}



import os

import yaml


class ConfigReader:
    @classmethod
    def read(cls, filepath):
        config = {}
        if os.path.exists(filepath):
            with open(filepath) as f:
                config = yaml.safe_load(f)
        return config



from optinist.api.dataclass.dataclass import ImageData, RoiData
from optinist.api.nwb.nwb import NWBDATASET


def caiman_mc(
    image: ImageData, output_dir: str, params: dict = None
) -> dict(mc_images=ImageData):
    import numpy as np
    from caiman import load_memmap, save_memmap, stop_server
    from caiman.base.rois import extract_binary_masks_from_structural_channel
    from caiman.cluster import setup_cluster
    from caiman.motion_correction import MotionCorrect
    from caiman.source_extraction.cnmf.params import CNMFParams

    opts = CNMFParams()

    if params is not None:
        opts.change_params(params_dict=params)

    c, dview, n_processes = setup_cluster(
        backend="local", n_processes=None, single_thread=True
    )

    mc = MotionCorrect(image.path, dview=dview, **opts.get_group("motion"))

    mc.motion_correct(save_movie=True)
    border_to_0 = 0 if mc.border_nan == "copy" else mc.border_to_0

    # memory mapping
    fname_new = save_memmap(
        mc.mmap_file, base_name="memmap_", order="C", border_to_0=border_to_0
    )

    stop_server(dview=dview)

    # now load the file
    Yr, dims, T = load_memmap(fname_new)

    images = np.array(Yr.T.reshape((T,) + dims, order="F"))

    meanImg = images.mean(axis=0)
    rois = (
        extract_binary_masks_from_structural_channel(
            meanImg, gSig=7, expand_method="dilation"
        )[0]
        .reshape(meanImg.shape[0], meanImg.shape[1], -1)
        .transpose(2, 0, 1)
    )

    rois = rois.astype(np.float)

    for i, _ in enumerate(rois):
        rois[i] *= i + 1

    rois = np.nanmax(rois, axis=0)
    rois[rois == 0] = np.nan

    xy_trans_data = (
        (np.array(mc.x_shifts_els), np.array(mc.y_shifts_els))
        if params["pw_rigid"]
        else np.array(mc.shifts_rig)
    )

    mc_images = ImageData(images, output_dir=output_dir, file_name="mc_images")

    nwbfile = {}
    nwbfile[NWBDATASET.MOTION_CORRECTION] = {
        "caiman_mc": {
            "mc_data": mc_images,
            "xy_trans_data": xy_trans_data,
        }
    }

    info = {
        "mc_images": mc_images,
        "meanImg": ImageData(meanImg, output_dir=output_dir, file_name="meanImg"),
        "rois": RoiData(rois, output_dir=output_dir, file_name="rois"),
        "nwbfile": nwbfile,
    }

    return info



import yaml

from optinist.api.utils.filepath_creater import create_directory, join_filepath


class ConfigWriter:
    @classmethod
    def write(cls, dirname, filename, config):
        create_directory(dirname)

        with open(join_filepath([dirname, filename]), "w") as f:
            yaml.dump(config, f)



import numpy as np
import scipy.sparse
import skimage

from optinist.wrappers.lccd_wrapper.lccd_python import utils


def oval_filter(label_mat, sparse=False):
    """ovel_filter functionality.
    Let label_mat has shape(X, Y),
    returned array has shape (X x Y, number of cells)

    when sparse=True, this method took 137ms
    when sparse=False, this method took 325ms,
    sparse=True gave 2.5x speedup.
    spatial size was 506 x 506, number of cells after filtering was 375.
    """
    if sparse:
        rows, data = [], []
    else:
        rois = []
    regionprops = skimage.measure.regionprops(label_mat)
    for regionprop in regionprops:
        area_ratio = (
            regionprop.axis_major_length
            * regionprop.axis_minor_length
            * np.pi
            / 4
            / regionprop.area
        )
        eccen = regionprop.eccentricity
        if area_ratio < 1.8 and eccen < 0.99:  # TODO Make configurable
            roi = (label_mat == regionprop.label).astype(np.uint8)
            if sparse:
                datum, row = utils.array_to_lil_row(roi.ravel())
                data.append(datum)
                rows.append(row)
            else:
                roi = roi.reshape(-1, 1)
                rois.append(roi)

    if sparse:
        if len(data) == 0:
            raise ValueError("No roi region found. Revise config and input.")
        lil_arr = scipy.sparse.lil_matrix(
            (len(data), label_mat.shape[0] * label_mat.shape[1]), dtype=data[0][0].dtype
        )
        lil_arr.data = np.array(data, dtype="object")
        lil_arr.rows = np.array(rows, dtype="object")
        csr_arr = scipy.sparse.csr_matrix(lil_arr)
        return csr_arr.T  # By transpose, csr_matrix is converted to csc_matrix.
    if len(rois) == 0:
        raise ValueError("No roi region found. Revise config and input.")
    rois = np.concatenate(rois, 1)
    return rois



import time

import numpy as np
from scipy import stats

from optinist.api.nwb.nwb import NWBDATASET


def masks_and_traces(ops, stat_manual, stat_orig):
    """main extraction function
    inputs: ops and stat
    creates cell and neuropil masks and extracts traces
    returns: F (ROIs x time), Fneu (ROIs x time), F_chan2, Fneu_chan2, ops, stat
    F_chan2 and Fneu_chan2 will be empty if no second channel
    """
    from suite2p import detection, extraction

    if "aspect" in ops:
        dy, dx = int(ops["aspect"] * 10), 10
    else:
        d0 = ops["diameter"]
        dy, dx = (d0, d0) if isinstance(d0, int) else d0
    t0 = time.time()
    # Concatenate stat so a good neuropil function can be formed
    stat_all = stat_orig
    for n in range(len(stat_manual)):
        stat_all.append(stat_manual[n])
    stat_all = detection.stats.roi_stats(stat_all, dy, dx, ops["Ly"], ops["Lx"])

    cell_masks = [
        extraction.masks.create_cell_mask(
            stat, Ly=ops["Ly"], Lx=ops["Lx"], allow_overlap=ops["allow_overlap"]
        )
        for stat in stat_all
    ]
    cell_pix = extraction.masks.create_cell_pix(stat_all, Ly=ops["Ly"], Lx=ops["Lx"])
    manual_roi_stats = stat_all[-len(stat_manual) :]
    manual_cell_masks = cell_masks[-len(stat_manual) :]

    manual_neuropil_masks = extraction.masks.create_neuropil_masks(
        ypixs=[stat["ypix"] for stat in manual_roi_stats],
        xpixs=[stat["xpix"] for stat in manual_roi_stats],
        cell_pix=cell_pix,
        inner_neuropil_radius=ops["inner_neuropil_radius"],
        min_neuropil_pixels=ops["min_neuropil_pixels"],
    )
    print("Masks made in %0.2f sec." % (time.time() - t0))

    F, Fneu, F_chan2, Fneu_chan2, ops = extraction.extract_traces_from_masks(
        ops, manual_cell_masks, manual_neuropil_masks
    )

    # compute activity statistics for classifier
    npix = np.array([stat_orig[n]["npix"] for n in range(len(stat_orig))]).astype(
        "float32"
    )
    for n in range(len(manual_roi_stats)):
        manual_roi_stats[n]["npix_norm"] = manual_roi_stats[n]["npix"] / np.mean(
            npix[:100]
        )  # What if there are less than 100 cells?
        manual_roi_stats[n]["compact"] = 1
        manual_roi_stats[n]["footprint"] = 2
        manual_roi_stats[n]["manual"] = 1  # Add manual key

    # subtract neuropil and compute skew, std from F
    dF = F - ops["neucoeff"] * Fneu
    sk = stats.skew(dF, axis=1)
    sd = np.std(dF, axis=1)

    for n in range(F.shape[0]):
        manual_roi_stats[n]["skew"] = sk[n]
        manual_roi_stats[n]["std"] = sd[n]
        manual_roi_stats[n]["med"] = [
            np.mean(manual_roi_stats[n]["ypix"]),
            np.mean(manual_roi_stats[n]["xpix"]),
        ]

    dF = extraction.preprocess(
        F=dF,
        baseline=ops["baseline"],
        win_baseline=ops["win_baseline"],
        sig_baseline=ops["sig_baseline"],
        fs=ops["fs"],
        prctile_baseline=ops["prctile_baseline"],
    )

    spks = extraction.dcnv.oasis(
        F=dF, batch_size=ops["batch_size"], tau=ops["tau"], fs=ops["fs"]
    )

    return F, Fneu, F_chan2, Fneu_chan2, spks, ops, manual_roi_stats


def get_stat0_add_roi(ops, posx, posy, sizex, sizey):
    posx, posy, sizex, sizey = round(posx), round(posy), round(sizex), round(sizey)
    xrange = (np.arange(-1 * int(sizex), 1) + int(posx)).astype(np.int32)
    yrange = (np.arange(-1 * int(sizey), 1) + int(posy)).astype(np.int32)
    xrange += int(np.floor(sizex / 2))
    yrange += int(np.floor(sizey / 2))

    ellipse = np.zeros((yrange.size, xrange.size), np.bool)
    x, y = np.meshgrid(np.arange(0, xrange.size, 1), np.arange(0, yrange.size, 1))
    ellipse = (
        (y - sizey / 2) ** 2 / (sizey / 2) ** 2
        + (x - sizex / 2) ** 2 / (sizex / 2) ** 2
    ) <= 1

    ellipse = ellipse[:, np.logical_and(xrange >= 0, xrange < ops["Lx"])]
    xrange = xrange[np.logical_and(xrange >= 0, xrange < ops["Lx"])]
    ellipse = ellipse[np.logical_and(yrange >= 0, yrange < ops["Ly"]), :]
    yrange = yrange[np.logical_and(yrange >= 0, yrange < ops["Ly"])]

    med = [sizex, sizey]
    x, y = np.meshgrid(xrange, yrange)
    ypix = y[ellipse].flatten()
    xpix = x[ellipse].flatten()
    lam = np.ones(ypix.shape)

    return [{"ypix": ypix, "xpix": xpix, "lam": lam, "npix": ypix.size, "med": med}]


def set_nwbfile(ops):
    stat = ops.get("stat")
    iscell = ops.get("iscell")
    F = ops.get("F")
    Fneu = ops.get("Fneu")

    roi_list = []
    for i in range(len(stat)):
        kargs = {}
        kargs["pixel_mask"] = np.array(
            [stat[i]["ypix"], stat[i]["xpix"], stat[i]["lam"]]
        ).T
        roi_list.append(kargs)
    nwbfile = {}

    nwbfile[NWBDATASET.ROI] = {"roi_list": roi_list}

    # iscellを追加
    nwbfile[NWBDATASET.COLUMN] = {
        "roi_column": {
            "name": "iscell",
            "discription": "two columns - iscell & probcell",
            "data": iscell,
        }
    }

    # Fluorenceを追加
    nwbfile[NWBDATASET.FLUORESCENCE] = {}
    for name, data in zip(["Fluorescence", "Neuropil"], [F, Fneu]):
        nwbfile[NWBDATASET.FLUORESCENCE][name] = {
            "table_name": name,
            "region": list(range(len(data))),
            "name": name,
            "data": data,
            "unit": "lumens",
            "rate": ops["fs"],
        }

    # NWB追加
    nwbfile[NWBDATASET.POSTPROCESS] = {
        "add_roi": ops.get("add_roi", []),
        "delete_roi": ops.get("delete_roi", []),
        "merge_roi": ops.get("merge_roi", []),
    }

    return nwbfile



from optinist.api.dataclass.dataclass import BarData, BehaviorData, FluoData, IscellData
from optinist.api.nwb.nwb import NWBDATASET
from optinist.wrappers.optinist_wrapper.utils import standard_norm


def SVM(
    neural_data: FluoData,
    behaviors_data: BehaviorData,
    output_dir: str,
    iscell: IscellData = None,
    params: dict = None,
) -> dict():
    # modules specific to function
    import numpy as np
    from sklearn import svm
    from sklearn.model_selection import GridSearchCV, StratifiedKFold

    neural_data = neural_data.data
    behaviors_data = behaviors_data.data

    # data shold be time x component matrix
    if params["transpose_x"]:
        X = neural_data.transpose()
    else:
        X = neural_data

    if params["transpose_y"]:
        Y = behaviors_data.transpose()
    else:
        Y = behaviors_data

    assert (
        X.shape[0] == Y.shape[0]
    ), f"""
        neural_data and behaviors_data is not same dimension,
        neural.shape{X.shape}, behavior.shape{Y.shape}"""

    if iscell is not None:
        iscell = iscell.data
        ind = np.where(iscell > 0)[0]
        X = X[:, ind]

    Y = Y[:, params["target_index"]].reshape(-1, 1)

    # preprocessing
    tX = standard_norm(X, params["standard_x_mean"], params["standard_x_std"])

    hp = params["SVC"]

    # SVM determination of hyper parameters if needed
    gs_clf = []
    if params["use_grid_search"]:
        param_grid = [params["grid_search"]["param_grid"]]
        gs_clf = GridSearchCV(
            svm.SVC(), param_grid=param_grid, **params["grid_search"]["CV"]
        )

        gs_clf.fit(tX, Y)

        # insert best parameter to hp dictionary
        keys = list(gs_clf.best_params_.keys())
        for i in range(len(keys)):
            hp[keys[i]] = gs_clf.best_params_[keys[i]]

    # cross validation of SVM using best grid search paraneters
    skf = StratifiedKFold(**params["CV"])

    score = []
    classifier = []
    for train_index, test_index in skf.split(tX, Y):
        clf = svm.SVC(**hp)

        if tX.shape[0] == 1:
            clf.fit(tX[train_index].reshape(-1, 1), Y[train_index])
            score.append(clf.score(tX[test_index].reshape(-1, 1), Y[test_index]))
        else:
            clf.fit(tX[train_index, :], Y[train_index])
            score.append(clf.score(tX[test_index, :], Y[test_index]))

        classifier.append(clf)

    # NWB追加
    nwbfile = {}
    nwbfile[NWBDATASET.POSTPROCESS] = {
        "score": score,
    }

    info = {"score": BarData(score, file_name="score"), "nwbfile": nwbfile}

    return info



import os

import numpy as np

from optinist.api.dataclass.base import BaseData


class LccdData(BaseData):
    def __init__(self, data, file_name="lccd"):
        super().__init__(file_name)
        self.data = data

    def save_json(self, json_dir):
        self.json_path = os.path.join(json_dir, f"{self.file_name}.npy")
        np.save(self.json_path, self.data)



from optinist.api.dataclass.base import BaseData
from optinist.api.utils.filepath_creater import join_filepath


class HTMLData(BaseData):
    def __init__(self, data, file_name="html"):
        super().__init__(file_name)
        self.data = data

    def save_json(self, json_dir):
        self.json_path = join_filepath([json_dir, f"{self.file_name}.html"])

        with open(self.json_path, "w") as f:
            f.write(self.data)



# flake8: noqa
# Exclude from lint for the following reason
# This file is executed by snakemake and cause the following lint errors
# - E402: sys.path.append is required to import optinist modules
# - F821: do not import snakemake
import sys

from const import OPTINIST_DIRPATH

sys.path.append(OPTINIST_DIRPATH)

from optinist.api.edit_ROI import EditRoiUtils

if __name__ == "__main__":
    config = snakemake.config
    EditRoiUtils.excute(config)



import numpy as np

from optinist.api.dataclass.dataclass import FluoData, ImageData, LccdData, RoiData
from optinist.api.nwb.nwb import NWBDATASET


def lccd_detect(
    mc_images: ImageData, output_dir: str, params: dict = None
) -> dict(fluorescence=FluoData, cell_roi=RoiData):
    from optinist.wrappers.lccd_wrapper.lccd_python.lccd import LCCD

    print("params: ", params)
    lccd = LCCD(params)
    D = LoadData(mc_images)
    assert len(D.shape) == 3, "input array should have dimensions (width, height, time)"
    roi = lccd.apply(D)

    dff_f0_frames = params["dff"]["f0_frames"]
    dff_f0_percentile = params["dff"]["f0_percentile"]
    num_cell = roi.shape[1]
    num_frames = D.shape[2]
    is_cell = np.ones(num_cell, dtype=bool)

    reshapedD = D.reshape([D.shape[0] * D.shape[1], D.shape[2]])
    timeseries = np.zeros([num_cell, num_frames])
    roi_list = []

    for i in range(num_cell):
        roi_list.append((roi[:, i].reshape([D.shape[0], D.shape[1]])) * (i + 1))
        timeseries[i, :] = np.mean(reshapedD[roi[:, i] > 0, :], axis=0)

    im = np.stack(roi_list)
    im[im == 0] = np.nan

    timeseries_dff = np.ones([num_cell, num_frames]) * np.nan
    for i in range(num_cell):
        for k in range(num_frames):
            if (k - dff_f0_frames >= 0) and (k + dff_f0_frames < num_frames):
                f0 = np.percentile(
                    timeseries[i, k - dff_f0_frames : k + dff_f0_frames],
                    dff_f0_percentile,
                )
                timeseries_dff[i, k] = (timeseries[i, k] - f0) / f0

    nwbfile = {}

    roi_list = [{"image_mask": roi[:, i].reshape(D.shape[:2])} for i in range(num_cell)]
    nwbfile[NWBDATASET.ROI] = {"roi_list": roi_list}

    nwbfile[NWBDATASET.COLUMN] = {
        "roi_column": {
            "name": "iscell",
            "discription": "two columns - iscell & probcell",
            "data": is_cell,
        }
    }

    nwbfile[NWBDATASET.FLUORESCENCE] = {}
    nwbfile[NWBDATASET.FLUORESCENCE]["Fluorescence"] = {
        "table_name": "Fluorescence",
        "region": list(range(len(timeseries))),
        "name": "Fluorescence",
        "data": timeseries,
        "unit": "lumens",
    }

    lccd_data = {}
    lccd_data["images"] = D
    lccd_data["roi"] = roi
    lccd_data["is_cell"] = is_cell

    info = {
        "lccd": LccdData(lccd_data),
        "cell_roi": RoiData(
            np.nanmax(im, axis=0), output_dir=output_dir, file_name="cell_roi"
        ),
        "fluorescence": FluoData(timeseries, file_name="fluorescence"),
        "dff": FluoData(timeseries_dff, file_name="dff"),
        "nwbfile": nwbfile,
    }

    return info


def LoadData(mc_images):
    from PIL import Image

    img_pile = Image.open(mc_images.path[0])
    num_page = img_pile.n_frames

    imgs = np.zeros((img_pile.size[0], img_pile.size[1], num_page))
    for i in range(num_page):
        img_pile.seek(i)
        imgs[:, :, i] = np.asarray(img_pile)
    img_pile.close()

    maxval = np.max(imgs)
    minval = np.min(imgs)
    imgs = (imgs - minval) / (maxval - minval)

    return imgs



import os

_DEFAULT_DIR = "/tmp/optinist"
_ENV_DIR = os.environ.get("OPTINIST_DIR")


class DIRPATH:
    OPTINIST_DIR = _DEFAULT_DIR if _ENV_DIR is None else _ENV_DIR
    INPUT_DIR = f"{OPTINIST_DIR}/input"
    OUTPUT_DIR = f"{OPTINIST_DIR}/output"

    CONDAYML_DIR = f"{os.path.dirname(os.path.dirname(__file__))}/conda"
    CONDAENV_DIR = (
        f"{os.path.dirname(os.path.dirname(os.path.dirname(__file__)))}/conda"
    )

    ROOT_DIR = os.path.dirname(os.path.dirname(__file__))
    CONFIG_DIR = f"{ROOT_DIR}/config"

    if not os.path.exists(INPUT_DIR):
        os.makedirs(INPUT_DIR)
    assert os.path.exists(INPUT_DIR)

    if not os.path.exists(OUTPUT_DIR):
        os.makedirs(OUTPUT_DIR)
    assert os.path.exists(OUTPUT_DIR)

    SNAKEMAKE_FILEPATH = f"{ROOT_DIR}/Snakefile"
    EXPERIMENT_YML = "experiment.yaml"
    SNAKEMAKE_CONFIG_YML = "config.yaml"



import os
import shutil


def join_filepath(path_list):
    if isinstance(path_list, str):
        return path_list
    elif isinstance(path_list, list):
        return "/".join(path_list)
    assert False, "Path is not list"


def create_filepath(dirname, filename):
    create_directory(dirname)

    return join_filepath([dirname, filename])


def get_pickle_file(unique_id, node_id, algo_name):
    return join_filepath([unique_id, node_id, f"{algo_name}.pkl"])


def create_directory(dirpath, delete_dir=False):
    if delete_dir and os.path.exists(dirpath):
        shutil.rmtree(dirpath)

    if not os.path.exists(dirpath):
        os.makedirs(dirpath)



from optinist.wrappers.suite2p_wrapper.file_convert import suite2p_file_convert
from optinist.wrappers.suite2p_wrapper.registration import suite2p_registration
from optinist.wrappers.suite2p_wrapper.roi import suite2p_roi
from optinist.wrappers.suite2p_wrapper.spike_deconv import suite2p_spike_deconv

suite2p_wrapper_dict = {
    "suite2p": {
        "suite2p_file_convert": {
            "function": suite2p_file_convert,
            "conda_name": "suite2p",
            "conda_yaml": "suite2p_env.yaml",
        },
        "suite2p_registration": {
            "function": suite2p_registration,
            "conda_name": "suite2p",
            "conda_yaml": "suite2p_env.yaml",
        },
        "suite2p_roi": {
            "function": suite2p_roi,
            "conda_name": "suite2p",
            "conda_yaml": "suite2p_env.yaml",
        },
        "suite2p_spike_deconv": {
            "function": suite2p_spike_deconv,
            "conda_name": "suite2p",
            "conda_yaml": "suite2p_env.yaml",
        },
    }
}



# flake8: noqa
# Exclude from lint for the following reason
# This file is executed by snakemake and cause the following lint errors
# - E402: sys.path.append is required to import optinist modules
# - F821: do not import snakemake
import sys

from const import OPTINIST_DIRPATH

sys.path.append(OPTINIST_DIRPATH)

from optinist.api.dir_path import DIRPATH
from optinist.api.rules.runner import Runner
from optinist.api.snakemake.snakemake_reader import RuleConfigReader
from optinist.api.utils.filepath_creater import join_filepath

if __name__ == "__main__":
    last_output = [
        join_filepath([DIRPATH.OUTPUT_DIR, x]) for x in snakemake.config["last_output"]
    ]

    rule_config = RuleConfigReader.read(snakemake.params.name)

    rule_config.input = snakemake.input
    rule_config.output = snakemake.output[0]

    Runner.run(rule_config, last_output)



import argparse
import os

import uvicorn
from fastapi import FastAPI, Request
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from starlette.middleware.cors import CORSMiddleware

from optinist.api.config.config_reader import ConfigReader
from optinist.api.dir_path import DIRPATH as OPTINIST_DIRPATH
from optinist.api.utils.filepath_creater import join_filepath
from optinist.routers import algolist, experiment, files, hdf5, outputs, params, run

app = FastAPI(docs_url="/docs", openapi_url="/openapi")
app.include_router(algolist.router)
app.include_router(files.router)
app.include_router(outputs.router)
app.include_router(params.router)
app.include_router(run.router)
app.include_router(hdf5.router)
app.include_router(experiment.router)

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

if "OPTINIST_DEV_ROOT_DIR" in os.environ:
    FRONTEND_DIRPATH = os.environ["OPTINIST_DEV_ROOT_DIR"] + "/frontend"
else:
    FRONTEND_DIRPATH = OPTINIST_DIRPATH.ROOT_DIR + "/frontend"

app.mount(
    "/static",
    StaticFiles(directory=f"{FRONTEND_DIRPATH}/build/static"),
    name="static",
)

templates = Jinja2Templates(directory=f"{FRONTEND_DIRPATH}/build")


@app.get("/")
async def root(request: Request):
    return templates.TemplateResponse("index.html", {"request": request})


def main(develop_mode: bool = False):
    parser = argparse.ArgumentParser()
    parser.add_argument("--host", type=str, default="127.0.0.1")
    parser.add_argument("--port", type=int, default=8000)
    parser.add_argument("--reload", action="store_true")
    args = parser.parse_args()

    # set fastapi@uvicorn logging config.
    logging_config_path = join_filepath(
        [OPTINIST_DIRPATH.ROOT_DIR, "app_config", "logging.yaml"]
    )
    logging_config = ConfigReader.read(logging_config_path)
    fastapi_logging_config = uvicorn.config.LOGGING_CONFIG
    fastapi_logging_config["formatters"]["default"]["fmt"] = logging_config[
        "fastapi_logging_config"
    ]["default_fmt"]
    fastapi_logging_config["formatters"]["access"]["fmt"] = logging_config[
        "fastapi_logging_config"
    ]["access_fmt"]

    if develop_mode:
        reload_options = {"reload_dirs": ["optinist"]} if args.reload else {}
        uvicorn.run(
            "optinist.__main_unit__:app",
            host=args.host,
            port=args.port,
            log_config=fastapi_logging_config,
            reload=args.reload,
            **reload_options,
        )
    else:
        uvicorn.run(
            "optinist.__main_unit__:app",
            host=args.host,
            port=args.port,
            log_config=fastapi_logging_config,
            reload=False,
        )



#  decoding neural activity by GLM
#  input:  A:matrix[num_cell x timeseries ]   B:timeseries(behavior)[1 x timeseries]
#  Generalized linear modeling using statsmodels
#
#  https://www.statsmodels.org/stable/glm.html

from optinist.api.dataclass.dataclass import (
    BarData,
    BehaviorData,
    FluoData,
    HTMLData,
    IscellData,
    ScatterData,
)
from optinist.api.nwb.nwb import NWBDATASET
from optinist.wrappers.optinist_wrapper.utils import standard_norm


def GLM(
    neural_data: FluoData,
    behaviors_data: BehaviorData,
    output_dir: str,
    iscell: IscellData = None,
    params: dict = None,
) -> dict():
    # modules specific to function
    import numpy as np
    import pandas as pd
    import statsmodels.api as sm

    neural_data = neural_data.data
    behaviors_data = behaviors_data.data

    # data shold be time x component matrix
    if params["transpose_x"]:
        X = neural_data.transpose()
    else:
        X = neural_data

    if params["transpose_y"]:
        Y = behaviors_data.transpose()
    else:
        Y = behaviors_data

    assert (
        X.shape[0] == Y.shape[0]
    ), f"""
        neural_data and behaviors_data is not same dimension,
        neural.shape{X.shape}, behavior.shape{Y.shape}"""

    if iscell is not None:
        iscell = iscell.data
        ind = np.where(iscell > 0)[0]
        X = X[:, ind]
        Y = Y[:, ind]

    Y = Y[:, params["target_index"]].reshape(-1, 1)

    # preprocessing
    tX = standard_norm(X, params["standard_x_mean"], params["standard_x_std"])
    tY = standard_norm(Y, params["standard_y_mean"], params["standard_y_std"])

    # calculate GLM
    tX = pd.DataFrame(tX)
    tY = pd.DataFrame(tY)

    if params["add_constant"]:
        tX = sm.add_constant(tX, prepend=False)

    # set family
    link = getattr(sm.genmod.families.links, params["link"])()  # noqa: F841
    family = eval(f"sm.families.{params['family']}(link=link)")

    # model fit
    model = sm.GLM(tY, tX, family=family, **params["GLM"])
    Res = model.fit()

    # NWB追加
    nwbfile = {}
    nwbfile[NWBDATASET.POSTPROCESS] = {
        "actual_predicted": np.array([Res._endog, Res.mu]).transpose(),
        "params": Res.params.values,
        "pvalues": Res.pvalues.values,
        "tvalues": Res.tvalues.values,  # z
        "aic": [Res.aic],
        "bic_llf": [Res.bic_llf],
        "llf": [Res.llf],  # log-Likelihood
        "pearson_chi2": [Res.pearson_chi2],
        "df_model": [Res.df_model],
        "df_resid": [Res.df_resid],
        "scale": [Res.scale],
        "mu": Res.mu,
        "endog": Res._endog,
    }

    # main results for plot
    # plot should be reconsidered --- what they should be!
    info = {
        "actual_predicted": ScatterData(
            np.array([Res._endog, Res.mu]).transpose(), file_name="actual_predicted"
        ),
        "params": BarData(Res.params.values, file_name="params"),
        "textout": HTMLData(Res.summary().as_html(), file_name="textout"),
        "nwbfile": nwbfile,
    }

    return info



import numpy as np
import pandas as pd

from optinist.api.dataclass.base import BaseData
from optinist.api.utils.filepath_creater import join_filepath
from optinist.api.utils.json_writer import JsonWriter


class HeatMapData(BaseData):
    def __init__(self, data, columns=None, file_name="heatmap"):
        super().__init__(file_name)
        self.data = data

        # indexを指定
        if columns is not None:
            self.columns = columns
        else:
            self.columns = np.arange(len(self.data[0]))

    def save_json(self, json_dir):
        self.json_path = join_filepath([json_dir, f"{self.file_name}.json"])
        df = pd.DataFrame(
            self.data,
            columns=self.columns,
        )
        JsonWriter.write_as_split(self.json_path, df)



import gc

import imageio
import numpy as np
import tifffile

from optinist.api.dataclass.base import BaseData
from optinist.api.dataclass.utils import create_images_list
from optinist.api.dir_path import DIRPATH
from optinist.api.utils.filepath_creater import create_directory, join_filepath
from optinist.api.utils.json_writer import JsonWriter


class ImageData(BaseData):
    def __init__(self, data, output_dir=DIRPATH.OUTPUT_DIR, file_name="image"):
        super().__init__(file_name)

        self.json_path = None

        if data is None:
            self.path = None
        elif isinstance(data, str):
            self.path = data
        elif isinstance(data, list) and isinstance(data[0], str):
            self.path = data
        else:
            _dir = join_filepath([output_dir, "tiff", file_name])
            create_directory(_dir)

            _path = join_filepath([_dir, f"{file_name}.tif"])
            tifffile.imsave(_path, data)
            self.path = [_path]

            del data
            gc.collect()

    @property
    def data(self):
        if isinstance(self.path, list):
            return np.concatenate([imageio.volread(p) for p in self.path])
        else:
            return np.array(imageio.volread(self.path))

    def save_json(self, json_dir):
        self.json_path = join_filepath([json_dir, f"{self.file_name}.json"])
        JsonWriter.write_as_split(self.json_path, create_images_list(self.data))


class RoiData(BaseData):
    def __init__(self, data, output_dir=DIRPATH.OUTPUT_DIR, file_name="roi"):
        super().__init__(file_name)

        images = create_images_list(data)

        _dir = join_filepath([output_dir, "tiff", file_name])
        create_directory(_dir)
        self.path = join_filepath([_dir, f"{file_name}.tif"])
        tifffile.imsave(self.path, images)

        del images, data
        gc.collect()

    @property
    def data(self):
        data = np.array(imageio.volread(self.path))
        if data.ndim == 3:
            return data[0]
        elif data.ndim == 2:
            return data

    def save_json(self, json_dir):
        self.json_path = join_filepath([json_dir, f"{self.file_name}.json"])
        JsonWriter.write_as_split(self.json_path, create_images_list(self.data))



from dataclasses import dataclass


@dataclass
class NWBDATASET:
    POSTPROCESS: str = "POSTPROCESS"
    TIMESERIES: str = "TIMESERIES"
    MOTION_CORRECTION: str = "MOCTION_CORRECTION"
    ROI: str = "ROI"
    COLUMN: str = "COLUMN"
    FLUORESCENCE: str = "FLUORESCENCE"
    BEHAVIOR: str = "BEHAVIOR"
    IMAGE_SERIES: str = "image_series"



import numpy as np
import pandas as pd

from optinist.api.dataclass.base import BaseData
from optinist.api.utils.filepath_creater import create_directory, join_filepath
from optinist.api.utils.json_writer import JsonWriter


class TimeSeriesData(BaseData):
    def __init__(
        self, data, std=None, index=None, cell_numbers=None, file_name="timeseries"
    ):
        super().__init__(file_name)

        assert data.ndim <= 2, "TimeSeries Dimension Error"

        if isinstance(data, str):
            self.data = pd.read_csv(data, header=None).values
        else:
            self.data = data

        if len(self.data.shape) == 1:
            self.data = self.data[np.newaxis, :]

        self.std = std

        # indexを指定
        if index is not None:
            self.index = index
        else:
            self.index = np.arange(len(self.data[0]))

        # cell番号を表示
        if cell_numbers is not None:
            self.cell_numbers = cell_numbers + 1
        else:
            self.cell_numbers = range(1, len(self.data) + 1)

    def save_json(self, json_dir):
        # timeseriesだけはdirを返す
        self.json_path = join_filepath([json_dir, self.file_name])
        create_directory(self.json_path, delete_dir=True)

        for i, cell_i in enumerate(self.cell_numbers):
            data = self.data[i]
            if self.std is not None:
                std = self.std[i]
                df = pd.DataFrame(
                    np.concatenate([data[:, np.newaxis], std[:, np.newaxis]], axis=1),
                    index=self.index,
                    columns=["data", "std"],
                )
            else:
                df = pd.DataFrame(data, index=self.index, columns=["data"])

            JsonWriter.write(join_filepath([self.json_path, f"{str(cell_i)}.json"]), df)


class FluoData(TimeSeriesData):
    def __init__(self, data, std=None, index=None, file_name="fluo"):
        super().__init__(
            data=data,
            std=std,
            index=index,
            file_name=file_name,
        )


class BehaviorData(TimeSeriesData):
    def __init__(self, data, std=None, index=None, file_name="behavior"):
        super().__init__(
            data=data,
            std=std,
            index=index,
            file_name=file_name,
        )



from optinist.api.config.config_writer import ConfigWriter
from optinist.api.dir_path import DIRPATH
from optinist.api.utils.filepath_creater import join_filepath


class SmkConfigWriter:
    @classmethod
    def write(cls, unique_id, flow_config):
        ConfigWriter.write(
            dirname=DIRPATH.ROOT_DIR,
            filename=DIRPATH.SNAKEMAKE_CONFIG_YML,
            config=flow_config,
        )

        ConfigWriter.write(
            dirname=join_filepath([DIRPATH.OUTPUT_DIR, unique_id]),
            filename=DIRPATH.SNAKEMAKE_CONFIG_YML,
            config=flow_config,
        )



from optinist.api.dataclass.bar import BarData
from optinist.api.dataclass.base import BaseData
from optinist.api.dataclass.caiman import CaimanCnmfData
from optinist.api.dataclass.csv import CsvData
from optinist.api.dataclass.heatmap import HeatMapData
from optinist.api.dataclass.html import HTMLData
from optinist.api.dataclass.image import ImageData, RoiData
from optinist.api.dataclass.iscell import IscellData
from optinist.api.dataclass.lccd import LccdData
from optinist.api.dataclass.nwb import NWBFile
from optinist.api.dataclass.scatter import ScatterData
from optinist.api.dataclass.suite2p import Suite2pData
from optinist.api.dataclass.timeseries import BehaviorData, FluoData, TimeSeriesData

__all__ = [
    "BarData",
    "BaseData",
    "CsvData",
    "CaimanCnmfData",
    "HeatMapData",
    "HTMLData",
    "ImageData",
    "RoiData",
    "IscellData",
    "LccdData",
    "NWBFile",
    "ScatterData",
    "Suite2pData",
    "BehaviorData",
    "FluoData",
    "TimeSeriesData",
]



from pynwb import get_class, load_namespaces
from pynwb.spec import NWBDatasetSpec, NWBGroupSpec, NWBNamespaceBuilder

name = "optinist"
ns_path = f"{name}.namespace.yaml"
ext_source = f"{name}.extensions.yaml"

# Now we define the data structures. We use `NWBDataInterface` as the base type,
# which is the most primitive type you are likely to use as a base. The name of the
# class is `CorticalSurface`, and it requires two matrices, `vertices` and
# `faces`.

postprocess = NWBGroupSpec(
    doc="postprocess",
    datasets=[
        NWBDatasetSpec(
            doc="data",
            shape=[
                (None,),
                (None, None),
                (None, None, None),
                (None, None, None, None),
            ],
            name="data",
            dtype="float",
        )
    ],
    neurodata_type_def="PostProcess",
    neurodata_type_inc="NWBDataInterface",
)

# Now we set up the builder and add this object

ns_builder = NWBNamespaceBuilder(f"{name} extensions", name, version="0.1.0")
ns_builder.add_spec(ext_source, postprocess)
ns_builder.export(ns_path)

load_namespaces(ns_path)
PostProcess = get_class("PostProcess", name)



import json

from optinist.routers.model import JsonTimeSeriesData, OutputData


class Reader:
    @classmethod
    def read(cls, filepath):
        with open(filepath, "r") as f:
            data = f.read()
        return data

    @classmethod
    def read_as_output(cls, filepath) -> OutputData:
        return OutputData(cls.read(filepath))


class JsonReader:
    @classmethod
    def read(cls, filepath):
        with open(filepath, "r") as f:
            json_data = json.load(f)
        return json_data

    @classmethod
    def read_as_output(cls, filepath) -> OutputData:
        json_data = cls.read(filepath)
        return OutputData(
            data=json_data["data"],
            columns=json_data["columns"],
            index=json_data["index"],
        )

    @classmethod
    def read_as_timeseries(cls, filepath) -> JsonTimeSeriesData:
        json_data = cls.read(filepath)
        return JsonTimeSeriesData(
            xrange=list(json_data["data"].keys()),
            data=json_data["data"],
            std=json_data["std"] if "std" in json_data else None,
        )



import gc


class BaseData:
    def __init__(self, file_name):
        self.file_name = file_name

    def save_json(self, json_dir):
        pass

    def __del__(self):
        del self
        gc.collect()



import copy
import gc
import os
import traceback

from optinist.api.nwb.nwb_creater import merge_nwbfile, save_nwb
from optinist.api.pickle.pickle_reader import PickleReader
from optinist.api.pickle.pickle_writer import PickleWriter
from optinist.api.snakemake.smk import Rule
from optinist.api.utils.filepath_creater import join_filepath
from optinist.wrappers import wrapper_dict


class Runner:
    @classmethod
    def run(cls, __rule: Rule, last_output):
        try:
            input_info = cls.read_input_info(__rule.input)

            cls.change_dict_key_exist(input_info, __rule)

            nwbfile = input_info["nwbfile"]

            # input_info
            for key in list(input_info):
                if key not in __rule.return_arg.values():
                    input_info.pop(key)

            # output_info
            output_info = cls.execute_function(
                __rule.path, __rule.params, os.path.dirname(__rule.output), input_info
            )

            # nwbfileの設定
            output_info["nwbfile"] = cls.save_func_nwb(
                f"{__rule.output.split('.')[0]}.nwb",
                __rule.type,
                nwbfile,
                output_info,
            )

            # 各関数での結果を保存
            PickleWriter.write(__rule.output, output_info)

            # NWB全体保存
            if __rule.output in last_output:
                # 全体の結果を保存する
                path = join_filepath(os.path.dirname(os.path.dirname(__rule.output)))
                path = join_filepath([path, f"whole_{__rule.type}.nwb"])
                cls.save_all_nwb(path, output_info["nwbfile"])

            print("output: ", __rule.output)

            del input_info, output_info
            gc.collect()

        except Exception as e:
            PickleWriter.write(
                __rule.output,
                list(traceback.TracebackException.from_exception(e).format())[-2:],
            )

    @classmethod
    def save_func_nwb(cls, save_path, name, nwbfile, output_info):
        if "nwbfile" in output_info:
            nwbfile[name] = output_info["nwbfile"]
            save_nwb(
                save_path,
                nwbfile["input"],
                output_info["nwbfile"],
            )
        return nwbfile

    @classmethod
    def save_all_nwb(cls, save_path, all_nwbfile):
        input_nwbfile = all_nwbfile["input"]
        all_nwbfile.pop("input")
        nwbfile = {}
        for x in all_nwbfile.values():
            nwbfile = merge_nwbfile(nwbfile, x)
        save_nwb(save_path, input_nwbfile, nwbfile)

    @classmethod
    def execute_function(cls, path, params, output_dir, input_info):
        wrapper = cls.dict2leaf(wrapper_dict, path.split("/"))
        func = copy.deepcopy(wrapper["function"])
        output_info = func(params=params, output_dir=output_dir, **input_info)
        del func
        gc.collect()

        return output_info

    @classmethod
    def change_dict_key_exist(cls, input_info, rule_config: Rule):
        for return_name, arg_name in rule_config.return_arg.items():
            if return_name in input_info:
                input_info[arg_name] = input_info.pop(return_name)

    @classmethod
    def read_input_info(cls, input_files):
        input_info = {}
        for filepath in input_files:
            load_data = PickleReader.read(filepath)
            input_info = dict(list(load_data.items()) + list(input_info.items()))
        return input_info

    @classmethod
    def dict2leaf(cls, root_dict: dict, path_list):
        path = path_list.pop(0)
        if len(path_list) > 0:
            return cls.dict2leaf(root_dict[path], path_list)
        else:
            return root_dict[path]



import numpy as np

from optinist.api.dataclass.dataclass import (
    HeatMapData,
    ImageData,
    IscellData,
    RoiData,
    ScatterData,
    TimeSeriesData,
)


def dummy_image2image(
    image: ImageData, params: dict = None
) -> dict(image2image=ImageData):
    """
    get image
    return image
    """
    info = {}
    info["image2image"] = ImageData(np.random.rand((300_0000)).reshape(300, 100, 100))
    return info


def dummy_image2time(
    image: ImageData, params: dict = None
) -> dict(image2time=TimeSeriesData):
    """
    get image
    return image
    """
    info = {}
    info["image2time"] = TimeSeriesData(np.random.rand((100000)).reshape(100, 1000))
    return info


def dummy_image2heat(
    image: ImageData, params: dict = None
) -> dict(image2heat=HeatMapData):
    """
    get image
    return image
    """
    info = {}
    info["image2heat"] = HeatMapData(np.random.rand((10000)).reshape(100, 100))
    return info


def dummy_time2time(
    timeseries: TimeSeriesData, params: dict = None
) -> dict(time2time=TimeSeriesData):
    """
    get image
    return image
    """
    info = {}
    info["time2time"] = TimeSeriesData(np.random.rand((10000)).reshape(10, 1000))
    return info


def dummy_image2image8time(
    image1: ImageData, params: dict = None
) -> dict(image=ImageData, timeseries=TimeSeriesData):
    """
    get image
    return image
    """
    info = {}
    info["image"] = ImageData(np.random.rand((100)).reshape(10, 10))
    info["timeseries"] = TimeSeriesData(np.random.rand((10000)).reshape(10, 1000))
    return info


def dummy_image8image2image8time(
    image1: ImageData, image2: ImageData, params: dict = None
) -> dict(image=ImageData, timeseries=TimeSeriesData):
    """
    get image
    return image
    """
    info = {}
    info["image"] = ImageData(np.random.rand((100)).reshape(10, 10), file_name="image")
    info["timeseries"] = TimeSeriesData(
        np.random.rand((10000)).reshape(10, 1000), file_name="timeseries"
    )
    return info


def dummy_time8image2image8time(
    timeseries: TimeSeriesData, image: ImageData, params: dict = None
) -> dict():
    """
    get image
    return image
    """
    info = {}
    info["image"] = ImageData(np.random.rand((100)).reshape(10, 10), file_name="image")
    info["timeseries"] = TimeSeriesData(
        np.random.rand((100)).reshape(10, 10), file_name="timeseries"
    )
    return info


def dummy_keyerror(image: ImageData, params: dict = None) -> dict():
    """
    get image
    return image
    """
    print(params["AAA"])
    info = {}
    return info


def dummy_typeerror(image: str, params: dict = None) -> dict():
    """
    get image
    return image
    """
    info = {}
    return info


def dummy_image2time8iscell(
    image1: ImageData, params: dict = None
) -> dict(timeseries=TimeSeriesData, iscell=IscellData):
    """
    get image
    return image
    """
    info = {}
    info["timeseries"] = TimeSeriesData(
        np.random.rand((100)).reshape(10, 10), file_name="image"
    )
    info["iscell"] = IscellData(
        np.random.rand((100)).reshape(10, 10), file_name="timeseries"
    )
    return info


def dummy_image2roi(image1: ImageData, params: dict = None) -> dict(roi=RoiData):
    """
    get image
    return image
    """
    roi_data = np.random.randint(low=1, high=1000, size=100_00).astype(float)
    import random

    null_data = np.array(random.sample(list(np.arange(100_00)), 90_00))
    roi_data[null_data] = np.nan
    roi_data = roi_data.reshape(100, 100)
    info = {}
    info["roi"] = RoiData(roi_data)
    return info


def dummy_image2image8roi(
    image1: ImageData, params: dict = None
) -> dict(image=ImageData, roi=RoiData):
    """
    get image
    return image
    """
    import random

    info = {}
    info["images"] = ImageData(np.random.rand((100_00)).reshape(1, 100, 100))
    roi_data = np.random.rand((100_00))
    null_data = np.array(random.sample(list(np.arange(100_00)), 90_00))
    roi_data[null_data] = np.nan
    roi_data = roi_data.reshape(100, 100)
    info["roi"] = RoiData(roi_data)
    return info


def dummy_image2image8roi8time8heat(
    image1: ImageData, params: dict = None
) -> dict(image=ImageData, roi=RoiData, timeseries=TimeSeriesData, heat=HeatMapData):
    """
    get image
    return image
    """
    import random

    info = {}
    info["images"] = ImageData(np.random.rand((100_00)).reshape(1, 100, 100))
    roi_data = np.random.rand((100_00))
    null_data = np.array(random.sample(list(np.arange(100_00)), 90_00))
    roi_data[null_data] = np.nan
    roi_data = roi_data.reshape(100, 100)
    info["roi"] = RoiData(roi_data)
    info["timeseries"] = TimeSeriesData(
        np.random.rand((100)).reshape(10, 10), file_name="image"
    )
    info["heat"] = HeatMapData(np.random.rand((10000)).reshape(100, 100))
    return info


def dummy_image2scatter(
    image: ImageData, params: dict = None
) -> dict(scatter=ScatterData):
    """
    get image
    return scatter
    """
    info = {}
    info["image2scatter"] = ScatterData(np.random.rand((1000)).reshape(500, 2))
    return info



import pickle

from optinist.api.utils.filepath_creater import create_directory, join_filepath


class PickleWriter:
    @classmethod
    def write(cls, pickle_path, info):
        # ファイル保存先
        dirpath = join_filepath(pickle_path.split("/")[:-1])
        create_directory(dirpath)
        with open(pickle_path, "wb") as f:
            pickle.dump(info, f)

    @classmethod
    def overwrite(cls, pickle_path, info):
        with open(pickle_path, "rb") as f:
            old_pkl = pickle.load(f)

        if type(old_pkl) == type(info) == dict:
            old_pkl.update(info)

            with open(pickle_path, "wb") as f:
                pickle.dump(old_pkl, f)



import uuid
from typing import Dict

from fastapi import APIRouter, BackgroundTasks

from optinist.api.workflow.workflow import Message, NodeItem, RunItem
from optinist.api.workflow.workflow_result import WorkflowResult
from optinist.api.workflow.workflow_runner import WorkflowRunner

router = APIRouter()


@router.post("/run", response_model=str, tags=["run"])
async def run(runItem: RunItem, background_tasks: BackgroundTasks):
    unique_id = str(uuid.uuid4())[:8]
    WorkflowRunner(unique_id, runItem).run_workflow(background_tasks)
    print("run snakemake")
    return unique_id


@router.post("/run/{uid}", response_model=str, tags=["run"])
async def run_id(uid: str, runItem: RunItem, background_tasks: BackgroundTasks):
    WorkflowRunner(uid, runItem).run_workflow(background_tasks)
    print("run snakemake")
    print("forcerun list: ", runItem.forceRunList)
    return uid


@router.post("/run/result/{uid}", response_model=Dict[str, Message], tags=["run"])
async def run_result(uid: str, nodeDict: NodeItem):
    return WorkflowResult(uid).get(nodeDict.pendingNodeIdList)



import os
from dataclasses import asdict
from datetime import datetime
from typing import Dict

from optinist.api.config.config_writer import ConfigWriter
from optinist.api.dir_path import DIRPATH
from optinist.api.experiment.experiment import ExptConfig, ExptFunction
from optinist.api.experiment.experiment_builder import ExptConfigBuilder
from optinist.api.experiment.experiment_reader import ExptConfigReader
from optinist.api.utils.filepath_creater import join_filepath
from optinist.api.workflow.workflow import Edge, Node


class ExptConfigWriter:
    def __init__(
        self,
        unique_id: str,
        name: str,
        nodeDict: Dict[str, Node],
        edgeDict: Dict[str, Edge],
    ) -> None:
        self.unique_id = unique_id
        self.name = name
        self.nodeDict = nodeDict
        self.edgeDict = edgeDict

        self.builder = ExptConfigBuilder()

    def write(self) -> None:
        expt_filepath = join_filepath(
            [DIRPATH.OUTPUT_DIR, self.unique_id, DIRPATH.EXPERIMENT_YML]
        )
        if os.path.exists(expt_filepath):
            expt_config = ExptConfigReader.read(expt_filepath)
            self.builder.set_config(expt_config)
            self.add_run_info()
        else:
            self.create_config()

        self.function_from_nodeDict()

        ConfigWriter.write(
            dirname=join_filepath([DIRPATH.OUTPUT_DIR, self.unique_id]),
            filename=DIRPATH.EXPERIMENT_YML,
            config=asdict(self.builder.build()),
        )

    def create_config(self) -> ExptConfig:
        return (
            self.builder.set_unique_id(self.unique_id)
            .set_name(self.name)
            .set_timestamp(datetime.now().strftime("%Y-%m-%d %H:%M:%S"))
            .set_nodeDict(self.nodeDict)
            .set_edgeDict(self.edgeDict)
            .build()
        )

    def add_run_info(self) -> ExptConfig:
        return (
            self.builder.set_timestamp(
                datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            )  # 時間を更新
            .set_nodeDict(self.nodeDict)
            .set_edgeDict(self.edgeDict)
            .build()
        )

    def function_from_nodeDict(self) -> ExptConfig:
        func_dict = {
            node.id: ExptFunction(
                unique_id=node.id,
                name=node.data.label,
                success="success" if node.data.type == "input" else "running",
                hasNWB=False,
            )
            for node in self.nodeDict.values()
        }
        return self.builder.set_function(func_dict).build()



from optinist.api.dataclass.dataclass import BarData, FluoData, IscellData, ScatterData
from optinist.api.nwb.nwb import NWBDATASET
from optinist.wrappers.optinist_wrapper.utils import standard_norm


def PCA(
    neural_data: FluoData,
    output_dir: str,
    iscell: IscellData = None,
    params: dict = None,
) -> dict():
    # modules specific to function
    import numpy as np
    from sklearn.decomposition import PCA

    neural_data = neural_data.data

    # data shold be time x component matrix
    if params["transpose"]:
        X = neural_data.transpose()
    else:
        X = neural_data

    if iscell is not None:
        iscell = iscell.data
        ind = np.where(iscell > 0)[0]
        X = X[:, ind]

    # # preprocessing
    tX = standard_norm(X, params["standard_mean"], params["standard_std"])

    # calculate PCA
    pca = PCA(**params["PCA"])
    proj_X = pca.fit_transform(tX)

    # NWB追加
    nwbfile = {}
    nwbfile[NWBDATASET.POSTPROCESS] = {
        "pca_projectedNd": proj_X,
        "components": pca.components_,
        "explained_variance": pca.explained_variance_,
        "explained_variance_ratio": pca.explained_variance_ratio_,
        "singular_values": pca.singular_values_,
        "mean": pca.mean_,
        "n_components": [pca.n_components_],
        "n_samples": [pca.n_samples_],
        "noise_variance": [pca.noise_variance_],
        "n_features_in": [pca.n_features_in_],
    }

    # import pdb; pdb.set_trace()
    info = {
        "explained_variance": BarData(pca.explained_variance_ratio_, file_name="evr"),
        "projectedNd": ScatterData(proj_X, file_name="projectedNd"),
        "contribution": BarData(
            pca.components_,
            index=[f"pca: {i}" for i in range(len(pca.components_))],
            file_name="contribution",
        ),
        "cumsum_contribution": BarData(
            np.cumsum(pca.components_, axis=0),
            index=[f"pca: {i}" for i in range(len(pca.components_))],
            file_name="cumsum_contribution",
        ),
        "nwbfile": nwbfile,
    }

    return info



import os

from optinist.api.dataclass.dataclass import ImageData, Suite2pData
from optinist.api.utils.filepath_creater import create_directory, join_filepath


def suite2p_file_convert(
    image: ImageData, output_dir: str, params: dict = None
) -> dict(ops=Suite2pData):
    from suite2p import default_ops, io

    print("start suite2_file_convert")

    data_path_list = []
    data_name_list = []
    for file_path in image.path:
        data_path_list.append(os.path.dirname(file_path))
        data_name_list.append(os.path.basename(file_path))

    print(data_path_list)
    print(data_name_list)
    # data pathと保存pathを指定
    db = {
        "data_path": data_path_list,
        "tiff_list": data_name_list,
        "save_path0": output_dir,
        "save_folder": "suite2p",
    }

    ops = {**default_ops(), **params, **db}

    # save folderを指定
    create_directory(join_filepath([ops["save_path0"], ops["save_folder"]]))

    # save ops.npy(parameter) and data.bin
    ops = io.tiff_to_binary(ops.copy())

    info = {
        "meanImg": ImageData(
            ops["meanImg"], output_dir=output_dir, file_name="meanImg"
        ),
        "ops": Suite2pData(ops, file_name="ops"),
    }

    return info



from optinist.wrappers.optinist_wrapper.basic_neural_analysis import (
    basic_neural_analysis_wrapper_dict,
)
from optinist.wrappers.optinist_wrapper.dimension_reduction import (
    dimension_reduction_wrapper_dict,
)
from optinist.wrappers.optinist_wrapper.neural_decoding import (
    neural_decoding_wrapper_dict,
)
from optinist.wrappers.optinist_wrapper.neural_population_analysis import (
    neural_population_analysis_wrapper_dict,
)

optinist_wrapper_dict = {
    "optinist": {
        "basic_neural_analysis": basic_neural_analysis_wrapper_dict,
        "dimension_reduction": dimension_reduction_wrapper_dict,
        "neural_population_analysis": neural_population_analysis_wrapper_dict,
        "neural_decoding": neural_decoding_wrapper_dict,
    }
}



import numpy as np
import pandas as pd

from optinist.api.dataclass.base import BaseData
from optinist.api.utils.filepath_creater import create_directory, join_filepath
from optinist.api.utils.json_writer import JsonWriter


class CsvData(BaseData):
    def __init__(self, data, params, file_name="csv"):
        super().__init__(file_name)

        if isinstance(data, str):
            self.data = pd.read_csv(data, header=None).values

            if params["transpose"]:
                self.data = self.data.T

            if params["setHeader"] is not None:
                header = params["setHeader"]
                self.data = self.data[header:]
        else:
            self.data = np.array(data)

        if self.data.ndim == 1:
            self.data = self.data[np.newaxis, :]

    def save_json(self, json_dir):
        # timeseriesだけはdirを返す
        self.json_path = join_filepath([json_dir, self.file_name])
        create_directory(self.json_path)

        for i, data in enumerate(self.data):
            JsonWriter.write_as_split(
                join_filepath([self.json_path, f"{str(i)}.json"]), data
            )



from optinist.wrappers.optinist_wrapper.neural_decoding.glm import GLM
from optinist.wrappers.optinist_wrapper.neural_decoding.lda import LDA
from optinist.wrappers.optinist_wrapper.neural_decoding.svm import SVM

neural_decoding_wrapper_dict = {
    "glm": {
        "function": GLM,
        "conda_yaml": "optinist_env.yaml",
    },
    "lda": {
        "function": LDA,
        "conda_yaml": "optinist_env.yaml",
    },
    "svm": {
        "function": SVM,
        "conda_yaml": "optinist_env.yaml",
    },
}



from optinist.api.dataclass.dataclass import CaimanCnmfData, FluoData, RoiData
from optinist.api.edit_ROI.utils import create_mask
from optinist.api.edit_ROI.wrappers.caiman_edit_roi.utils import set_nwbfile


def execute_add_ROI(node_dirpath, posx, posy, sizex, sizey):
    import numpy as np

    # load data
    cnmf_data = np.load(f"{node_dirpath}/caiman_cnmf.npy", allow_pickle=True).item()
    is_cell = cnmf_data.get("is_cell")
    im = cnmf_data.get("im")
    images = cnmf_data.get("images")
    add_roi = cnmf_data.get("add_roi", [])
    fluorescence = cnmf_data.get("fluorescence")

    # Create mask for new roi
    new_roi = create_mask(posx, posy, sizex, sizey, images.shape[1:])

    im = np.concatenate((im, new_roi[np.newaxis, :, :]), axis=0)
    is_cell = np.append(is_cell, True)

    # extract fluorescence of new ROI
    num_frames = images.shape[0]
    reshapeImages = images.reshape([images.shape[2] * images.shape[1], num_frames])
    new_fluorescence = np.zeros(num_frames)
    new_fluorescence = np.mean(reshapeImages[new_roi.reshape(-1, 1)[:, 0] > 0], axis=0)
    fluorescence = np.vstack([fluorescence, new_fluorescence])

    cell_roi = np.zeros(im.shape)
    num_rois = im.shape[0]
    for i in range(num_rois):
        cell_roi[i, :, :] = np.where(im[i, :, :] != 0, i + 1, np.nan)
    add_roi.append(num_rois)

    # save data
    cnmf_data["im"] = im
    cnmf_data["is_cell"] = is_cell
    cnmf_data["fluorescence"] = fluorescence
    cnmf_data["add_roi"] = add_roi

    info = {
        "fluorescence": FluoData(fluorescence, file_name="fluorescence"),
        "cell_roi": RoiData(
            np.nanmax(cell_roi[is_cell], axis=0),
            output_dir=node_dirpath,
            file_name="cell_roi",
        ),
        "cnmf_data": CaimanCnmfData(cnmf_data),
        "nwbfile": set_nwbfile(cnmf_data),
    }

    return info



from optinist.api.snakemake.smk import Rule, SmkParam


class RuleConfigReader:
    @classmethod
    def read(cls, rule):
        return Rule(
            input=rule["input"],
            return_arg=rule["return_arg"],
            params=rule["params"],
            output=rule["output"],
            type=rule["type"],
            nwbfile=rule["nwbfile"],
            hdf5Path=rule["hdf5Path"],
            path=rule["path"],
        )


class SmkParamReader:
    @classmethod
    def read(cls, params):
        return SmkParam(
            use_conda=params["use_conda"],
            cores=params["cores"],
            forceall=params["forceall"],
            forcetargets=params["forcetargets"],
            lock=params["lock"],
            forcerun=params["forcerun"] if "forcerun" in params else [],
        )



from optinist.api.dataclass.dataclass import (
    BarData,
    BehaviorData,
    FluoData,
    IscellData,
    ScatterData,
)
from optinist.api.nwb.nwb import NWBDATASET
from optinist.wrappers.optinist_wrapper.utils import standard_norm


def CCA(
    neural_data: FluoData,
    behaviors_data: BehaviorData,
    output_dir: str,
    iscell: IscellData = None,
    params: dict = None,
) -> dict():
    import numpy as np
    from sklearn.cross_decomposition import CCA

    neural_data = neural_data.data
    behaviors_data = behaviors_data.data

    # data shold be time x component matrix
    if params["transpose_x"]:
        X = neural_data.transpose()
    else:
        X = neural_data

    if params["transpose_y"]:
        Y = behaviors_data.transpose()
    else:
        Y = behaviors_data

    assert (
        X.shape[0] == Y.shape[0]
    ), f"""
        neural_data and behaviors_data is not same dimension,
        neural.shape{X.shape}, behavior.shape{Y.shape}"""

    if iscell is not None:
        iscell = iscell.data
        ind = np.where(iscell > 0)[0]
        X = X[:, ind]

    Y = Y[:, params["target_index"]].reshape(-1, 1)

    # preprocessing
    tX = standard_norm(X, params["standard_x_mean"], params["standard_x_std"])
    tY = standard_norm(Y, params["standard_y_mean"], params["standard_y_std"])

    # calculate CCA
    cca = CCA(**params["CCA"])
    projX, projY = cca.fit_transform(tX, tY)

    proj = np.concatenate([projX, projY], axis=1)

    # NWB追加
    nwbfile = {}
    nwbfile[NWBDATASET.POSTPROCESS] = {
        "projectedNd": proj,
        "x_weights": cca.x_weights_,  # singular vectors
        "y_weights": cca.y_weights_,
        "x_loadings_": cca.x_rotations_,
        "y_loadings_": cca.x_rotations_,
        "coef": cca.coef_,
        "n_iter_": cca.n_iter_,
        # 'n_features_in_': [cca.n_features_in_],
    }

    info = {
        "projectedNd": ScatterData(proj, file_name="projectedNd"),
        "coef": BarData(cca.coef_.flatten(), file_name="coef"),
        "nwbfile": nwbfile,
    }

    return info



from optinist.wrappers.optinist_wrapper.dimension_reduction.cca import CCA
from optinist.wrappers.optinist_wrapper.dimension_reduction.pca import PCA
from optinist.wrappers.optinist_wrapper.dimension_reduction.tsne import TSNE

dimension_reduction_wrapper_dict = {
    "cca": {
        "function": CCA,
        "conda_yaml": "optinist_env.yaml",
    },
    "pca": {
        "function": PCA,
        "conda_yaml": "optinist_env.yaml",
    },
    "tsne": {
        "function": TSNE,
        "conda_yaml": "optinist_env.yaml",
    },
}



import copy
import gc
import os
from dataclasses import dataclass
from glob import glob

from fastapi import HTTPException, status
from snakemake import snakemake

from optinist.api.config.config_reader import ConfigReader
from optinist.api.config.config_writer import ConfigWriter
from optinist.api.dataclass.base import BaseData
from optinist.api.dir_path import DIRPATH
from optinist.api.nwb.nwb_creater import overwrite_nwb
from optinist.api.pickle.pickle_reader import PickleReader
from optinist.api.pickle.pickle_writer import PickleWriter
from optinist.api.rules.runner import Runner
from optinist.api.utils.filepath_creater import join_filepath


@dataclass
class ACTION:
    EDIT_ROI: str = "edit_ROI"
    ADD: str = "add"
    MERGE: str = "merge"
    DELETE: str = "delete"


class EditROI:
    def __init__(self, action: str, filepath: str, params: dict) -> None:
        self.action = action
        self.filepath = filepath
        self.params = params
        self.algo = self.get_algo()
        self.node_dirpath = os.path.dirname(self.filepath)

    def get_algo(self):
        from optinist.api.edit_ROI import edit_roi_wrapper_dict

        filepath = self.filepath

        if (
            not os.path.exists(filepath)
            and os.path.commonpath([DIRPATH.OPTINIST_DIR, filepath])
            != DIRPATH.OPTINIST_DIR
        ):
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)
        if os.path.basename(filepath) != "cell_roi.json":
            raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST)

        algo_list = edit_roi_wrapper_dict.keys()

        algo = next((algo for algo in algo_list if algo in filepath), None)
        if not algo:
            raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST)
        return algo

    def excute(self):
        self.set_smk_config()

        snakemake(
            DIRPATH.SNAKEMAKE_FILEPATH,
            use_conda=True,
            cores=2,
            workdir=f"{os.path.dirname(DIRPATH.ROOT_DIR)}",
        )

    def set_smk_config(self):
        config = {
            "type": ACTION.EDIT_ROI,
            "action": self.action,
            "node_dirpath": self.node_dirpath,
            "algo": self.algo,
            "params": self.params,
        }
        ConfigWriter.write(
            dirname=DIRPATH.ROOT_DIR,
            filename=DIRPATH.SNAKEMAKE_CONFIG_YML,
            config=config,
        )


class EditRoiUtils:
    @classmethod
    def conda(cls, config):
        from optinist.api.edit_ROI import edit_roi_wrapper_dict

        algo = config["algo"]
        if "conda_yaml" in edit_roi_wrapper_dict[algo]:
            conda_yaml = edit_roi_wrapper_dict[algo]["conda_yaml"]
            return f"{DIRPATH.CONDAYML_DIR}/{conda_yaml}" if conda_yaml else None

        return None

    @classmethod
    def excute(cls, config):
        from optinist.api.edit_ROI import edit_roi_wrapper_dict

        algo = config["algo"]
        node_dirpath = config["node_dirpath"]
        action = config["action"]
        params = config["params"]

        func = copy.deepcopy(edit_roi_wrapper_dict[algo]["function"][action])
        output_info = func(node_dirpath, **params)
        del func
        gc.collect()

        workflow_dirpath = os.path.dirname(node_dirpath)
        smk_config_file = join_filepath(
            [workflow_dirpath, DIRPATH.SNAKEMAKE_CONFIG_YML]
        )
        smk_config = ConfigReader.read(smk_config_file)
        last_outputs = smk_config.get("last_output")
        pickle_files = glob(join_filepath([node_dirpath, "*.pkl"]))

        if len(pickle_files) > 0:
            func_name = os.path.splitext(os.path.basename(pickle_files[0]))[0]
            cls.__update_pickle_for_roi_edition(pickle_files[0], func_name, output_info)

            for last_output in last_outputs:
                last_output_path = join_filepath([DIRPATH.OUTPUT_DIR, last_output])
                last_output_info = cls.__update_pickle_for_roi_edition(
                    last_output_path, func_name, output_info
                )
                rule_type = os.path.splitext(os.path.basename(last_output_path))[0]
                whole_nwb_path = join_filepath(
                    [workflow_dirpath, f"whole_{rule_type}.nwb"]
                )

                Runner.save_all_nwb(whole_nwb_path, last_output_info["nwbfile"])

                del last_output_info
                gc.collect()

        for k, v in output_info.items():
            if isinstance(v, BaseData):
                v.save_json(node_dirpath)

            if k == "nwbfile":
                nwb_files = glob(join_filepath([node_dirpath, "*.nwb"]))
                if len(nwb_files) > 0:
                    overwrite_nwb(v, node_dirpath, os.path.basename(nwb_files[0]))

        del output_info
        gc.collect()

    @classmethod
    def __update_pickle_for_roi_edition(cls, filepath, func_name, new_output_info):
        output_info = PickleReader.read(filepath)
        for k, v in new_output_info.items():
            if k == "nwbfile":
                output_info[k][func_name] = v
            else:
                output_info[k] = v
        PickleWriter.overwrite(pickle_path=filepath, info=output_info)
        return output_info



from optinist.api.snakemake.smk import Rule


class RuleBuilder:
    def __init__(self) -> None:
        self._input = None
        self._return_arg = None
        self._params = None
        self._output = None
        self._type = None
        self._nwbfile = None
        self._hdf5Path = None
        self._path = None

    def set_input(self, input) -> "RuleBuilder":
        self._input = input
        return self

    def set_return_arg(self, return_arg) -> "RuleBuilder":
        self._return_arg = return_arg
        return self

    def set_params(self, params) -> "RuleBuilder":
        self._params = params
        return self

    def set_output(self, output) -> "RuleBuilder":
        self._output = output
        return self

    def set_type(self, type) -> "RuleBuilder":
        self._type = type
        return self

    def set_nwbfile(self, nwbfile) -> "RuleBuilder":
        self._nwbfile = nwbfile
        return self

    def set_hdf5Path(self, hdf5Path) -> "RuleBuilder":
        self._hdf5Path = hdf5Path
        return self

    def set_path(self, path) -> "RuleBuilder":
        self._path = path
        return self

    def build(self) -> Rule:
        return Rule(
            input=self._input,
            return_arg=self._return_arg,
            params=self._params,
            output=self._output,
            type=self._type,
            nwbfile=self._nwbfile,
            hdf5Path=self._hdf5Path,
            path=self._path,
        )



import os

import numpy as np

from optinist.api.dataclass.dataclass import FluoData, RoiData, Suite2pData
from optinist.api.edit_ROI.wrappers.suite2p_edit_roi.utils import (
    get_stat0_add_roi,
    masks_and_traces,
    set_nwbfile,
)


def execute_add_ROI(node_dirpath, posx, posy, sizex, sizey):
    from suite2p import ROI

    ops = np.load(os.path.join(node_dirpath, "suite2p.npy"), allow_pickle=True).item()
    iscell = ops.get("iscell")
    stat = ops.get("stat")
    stat0 = get_stat0_add_roi(ops, posx, posy, sizex, sizey)
    im = ops.get("im")
    add_roi = ops.get("add_roi", [])

    Fcell, Fneu, _, _, _, ops, manual_stat_cell = masks_and_traces(ops, stat0, stat)

    for n in range(len(manual_stat_cell)):
        array = np.expand_dims(
            ROI(
                ypix=manual_stat_cell[n]["ypix"],
                xpix=manual_stat_cell[n]["xpix"],
                lam=manual_stat_cell[n]["lam"],
                med=manual_stat_cell[n]["med"],
                do_crop=False,
            ).to_array(Ly=ops["Ly"], Lx=ops["Lx"]),
            axis=0,
        )
        id = np.nanmax(im) + 1
        array *= id
        add_roi.append(id)
        im = np.concatenate((im, array), axis=0)
    im[im == 0] = np.nan

    # Save fluorescence traces
    F_all = np.concatenate((ops["F"], Fcell), axis=0)
    Fneu_all = np.concatenate((ops["Fneu"], Fneu), axis=0)
    iscell = np.append(iscell, True)
    ops["stat"] = stat
    ops["F"] = F_all
    ops["Fneu"] = Fneu_all
    ops["iscell"] = iscell
    ops["im"] = im
    ops["add_roi"] = add_roi

    info = {
        "ops": Suite2pData(ops),
        "fluorescence": FluoData(ops["F"], file_name="fluorescence"),
        "all_roi": RoiData(
            np.nanmax(im, axis=0), output_dir=node_dirpath, file_name="all_roi"
        ),
        "cell_roi": RoiData(
            np.nanmax(im[iscell], axis=0), output_dir=node_dirpath, file_name="cell_roi"
        ),
        "nwbfile": set_nwbfile(ops),
    }
    return info



from typing import Dict

from optinist.api.snakemake.smk import Rule
from optinist.api.snakemake.smk_builder import RuleBuilder
from optinist.api.utils.filepath_creater import get_pickle_file
from optinist.api.workflow.workflow import Edge, Node, NodeType
from optinist.api.workflow.workflow_params import get_typecheck_params


class SmkRule:
    def __init__(
        self, unique_id: str, node: Node, edgeDict: Dict[str, Edge], nwbfile=None
    ) -> None:
        self._unique_id = unique_id
        self._node = node
        self._edgeDict = edgeDict
        self._nwbfile = nwbfile

        _return_name = self.get_return_name()

        _output_file = get_pickle_file(
            self._unique_id, self._node.id, self._node.data.label.split(".")[0]
        )

        self.builder = RuleBuilder()
        (
            self.builder.set_input(self._node.data.path)
            .set_return_arg(_return_name)
            .set_params(self._node.data.param)
            .set_output(_output_file)
            .set_nwbfile(self._nwbfile)
        )

    def image(self) -> Rule:
        return self.builder.set_type("image").build()

    def csv(self, nodeType="csv") -> Rule:
        return self.builder.set_type(nodeType).build()

    def hdf5(self) -> Rule:
        return (
            self.builder.set_type("hdf5").set_hdf5Path(self._node.data.hdf5Path).build()
        )

    def algo(self, nodeDict: Dict[str, Node]) -> Rule:
        algo_input = []
        return_arg_names = {}
        for edge in self._edgeDict.values():
            if self._node.id == edge.target:
                arg_name = edge.targetHandle.split("--")[1]

                sourceNode = nodeDict[edge.source]
                if sourceNode.type == NodeType.ALGO:
                    return_name = edge.sourceHandle.split("--")[1]
                    funcname = sourceNode.data.label
                else:
                    return_name = edge.sourceHandle.split("--")[0]
                    funcname = sourceNode.data.label.split(".")[0]

                algo_input.append(
                    get_pickle_file(self._unique_id, sourceNode.id, funcname)
                )

                return_arg_names[return_name] = arg_name

        params = get_typecheck_params(self._node.data.param, self._node.data.label)
        algo_output = get_pickle_file(
            self._unique_id, self._node.id, self._node.data.label
        )

        return (
            self.builder.set_input(algo_input)
            .set_return_arg(return_arg_names)
            .set_params(params)
            .set_output(algo_output)
            .set_path(self._node.data.path)
            .set_type(self._node.data.label)
            .build()
        )

    def get_return_name(self) -> str or None:
        for edge in self._edgeDict.values():
            if self._node.id == edge.source:
                return edge.sourceHandle.split("--")[0]
        return None



import numpy as np
import pandas as pd

from optinist.api.dataclass.base import BaseData
from optinist.api.utils.filepath_creater import join_filepath
from optinist.api.utils.json_writer import JsonWriter


class BarData(BaseData):
    def __init__(self, data, index=None, file_name="bar"):
        super().__init__(file_name)
        data = np.array(data)

        assert data.ndim <= 2, "Bar Dimension Error"

        if data.ndim == 1:
            data = data[np.newaxis]

        assert data.ndim == 2, "Bar Dimesion is not 2"

        self.data = data

        # indexを指定
        if index is not None:
            self.index = index
        else:
            self.index = np.arange(len(self.data))

    def save_json(self, json_dir):
        self.json_path = join_filepath([json_dir, f"{self.file_name}.json"])
        df = pd.DataFrame(
            self.data,
            index=self.index,
        )
        JsonWriter.write_as_split(self.json_path, df)



import os
from dataclasses import asdict
from glob import glob
from typing import Dict

from optinist.api.config.config_writer import ConfigWriter
from optinist.api.dataclass.dataclass import (
    BarData,
    BaseData,
    HeatMapData,
    HTMLData,
    ImageData,
    RoiData,
    ScatterData,
    TimeSeriesData,
)
from optinist.api.dir_path import DIRPATH
from optinist.api.experiment.experiment_reader import ExptConfigReader
from optinist.api.pickle.pickle_reader import PickleReader
from optinist.api.utils.filepath_creater import join_filepath
from optinist.api.workflow.workflow import Message, OutputPath, OutputType
from optinist.routers.fileIO.file_reader import Reader


class WorkflowResult:
    def __init__(self, unique_id):
        self.unique_id = unique_id
        self.workflow_dirpath = join_filepath(
            [
                DIRPATH.OUTPUT_DIR,
                self.unique_id,
            ]
        )
        self.expt_filepath = join_filepath(
            [self.workflow_dirpath, DIRPATH.EXPERIMENT_YML]
        )
        self.error_filepath = join_filepath([self.workflow_dirpath, "error.log"])

    def get(self, nodeIdList):
        results: Dict[str, Message] = {}
        for node_id in nodeIdList:
            if os.path.exists(self.error_filepath):
                error_message = Reader.read(self.error_filepath)
                if error_message != "":
                    results[node_id] = Message(
                        status="error",
                        message=error_message,
                    )

            glob_pickle_filepath = join_filepath(
                [self.workflow_dirpath, node_id, "*.pkl"]
            )
            for pickle_filepath in glob(glob_pickle_filepath):
                results[node_id] = NodeResult(
                    self.workflow_dirpath,
                    node_id,
                    pickle_filepath,
                ).get()
                self.has_nwb(node_id)

        self.has_nwb()

        return results

    def has_nwb(self, node_id=None):
        if node_id is None:
            nwb_filepath_list = glob(join_filepath([self.workflow_dirpath, "*.nwb"]))
        else:
            nwb_filepath_list = glob(
                join_filepath([self.workflow_dirpath, node_id, "*.nwb"])
            )

        for nwb_filepath in nwb_filepath_list:
            if os.path.exists(nwb_filepath):
                config = ExptConfigReader.read(self.expt_filepath)

                if node_id is None:
                    config.hasNWB = True
                else:
                    config.function[node_id].hasNWB = True

                ConfigWriter.write(
                    dirname=self.workflow_dirpath,
                    filename=DIRPATH.EXPERIMENT_YML,
                    config=asdict(config),
                )


class NodeResult:
    def __init__(self, workflow_dirpath, node_id, pickle_filepath):
        self.workflow_dirpath = workflow_dirpath
        self.node_id = node_id
        self.node_dirpath = join_filepath([self.workflow_dirpath, self.node_id])
        self.expt_filepath = join_filepath(
            [self.workflow_dirpath, DIRPATH.EXPERIMENT_YML]
        )

        pickle_filepath = pickle_filepath.replace("\\", "/")
        self.algo_name = os.path.splitext(os.path.basename(pickle_filepath))[0]
        self.info = PickleReader.read(pickle_filepath)

    def get(self):
        expt_config = ExptConfigReader.read(self.expt_filepath)
        if isinstance(self.info, (list, str)):
            expt_config.function[self.node_id].success = "error"
            message = self.error()
        else:
            expt_config.function[self.node_id].success = "success"
            message = self.success()

        ConfigWriter.write(
            dirname=self.workflow_dirpath,
            filename=DIRPATH.EXPERIMENT_YML,
            config=asdict(expt_config),
        )

        return message

    def success(self):
        return Message(
            status="success",
            message=f"{self.algo_name} success",
            outputPaths=self.outputPaths(),
        )

    def error(self):
        return Message(
            status="error",
            message=self.info if isinstance(self.info, str) else "\n".join(self.info),
        )

    def outputPaths(self):
        outputPaths: Dict[str, OutputPath] = {}
        for k, v in self.info.items():
            if isinstance(v, BaseData):
                v.save_json(self.node_dirpath)

            if isinstance(v, ImageData):
                outputPaths[k] = OutputPath(
                    path=v.json_path,
                    type=OutputType.IMAGE,
                    max_index=len(v.data) if v.data.ndim == 3 else 1,
                )
            elif isinstance(v, TimeSeriesData):
                outputPaths[k] = OutputPath(
                    path=v.json_path, type=OutputType.TIMESERIES, max_index=len(v.data)
                )
            elif isinstance(v, HeatMapData):
                outputPaths[k] = OutputPath(
                    path=v.json_path,
                    type=OutputType.HEATMAP,
                )
            elif isinstance(v, RoiData):
                outputPaths[k] = OutputPath(
                    path=v.json_path,
                    type=OutputType.ROI,
                )
            elif isinstance(v, ScatterData):
                outputPaths[k] = OutputPath(
                    path=v.json_path,
                    type=OutputType.SCATTER,
                )
            elif isinstance(v, BarData):
                outputPaths[k] = OutputPath(
                    path=v.json_path,
                    type=OutputType.BAR,
                )
            elif isinstance(v, HTMLData):
                outputPaths[k] = OutputPath(
                    path=v.json_path,
                    type=OutputType.HTML,
                )
            else:
                pass

        return outputPaths



from optinist.api.dataclass.base import BaseData
from optinist.api.utils.filepath_creater import join_filepath
from optinist.api.utils.json_writer import JsonWriter


class ScatterData(BaseData):
    def __init__(self, data, file_name="scatter"):
        super().__init__(file_name)

        assert data.ndim <= 2, "Scatter Dimension Error"

        self.data = data.T

    def save_json(self, json_dir):
        self.json_path = join_filepath([json_dir, f"{self.file_name}.json"])
        JsonWriter.write_as_split(self.json_path, self.data)



# flake8: noqa
# Exclude from lint for the following reason
# This file is executed by snakemake and cause the following lint errors
# - E402: sys.path.append is required to import optinist modules
# - F821: do not import snakemake
import sys

from const import OPTINIST_DIRPATH

sys.path.append(OPTINIST_DIRPATH)

from optinist.api.pickle.pickle_writer import PickleWriter
from optinist.api.rules.file_writer import FileWriter
from optinist.api.snakemake.snakemake_reader import RuleConfigReader
from optinist.routers.model import FILETYPE

if __name__ == "__main__":
    last_output = snakemake.config["last_output"]

    rule_config = RuleConfigReader.read(snakemake.params.name)
    if rule_config.type in [FILETYPE.IMAGE]:
        rule_config.input = snakemake.input
    elif rule_config.type in [FILETYPE.CSV, FILETYPE.BEHAVIOR, FILETYPE.HDF5]:
        rule_config.input = snakemake.input[0]

    rule_config.output = snakemake.output[0]

    if rule_config.type in [FILETYPE.CSV, FILETYPE.BEHAVIOR]:
        outputfile = FileWriter.csv(rule_config, rule_config.type)
        PickleWriter.write(rule_config.output, outputfile)
    elif rule_config.type == FILETYPE.IMAGE:
        outputfile = FileWriter.image(rule_config)
        PickleWriter.write(rule_config.output, outputfile)
    elif rule_config.type == FILETYPE.HDF5:
        outputfile = FileWriter.hdf5(rule_config)
        PickleWriter.write(rule_config.output, outputfile)



import copy

import numpy as np


def create_images_list(data):
    if len(data.shape) == 2:
        save_data = copy.deepcopy(data)
    elif len(data.shape) == 3:
        save_data = copy.deepcopy(data[:10])
    else:
        assert False, "data is error"

    if len(data.shape) == 2:
        data = data[np.newaxis, :, :]
        save_data = save_data[np.newaxis, :, :]

    images = []
    for _img in save_data:
        images.append(_img.tolist())

    return images



from typing import Tuple


def create_mask(x: int, y: int, width: int, height: int, dims: Tuple[int, int]):
    import numpy as np

    x, y, width, height = round(x), round(y), round(width), round(height)

    x_coords = np.arange(0, dims[0])
    y_coords = np.arange(0, dims[1])
    xx, yy = np.meshgrid(x_coords, y_coords)

    # Calculate the distance of each pixel from the center of the ellipse
    a = width / 2
    b = height / 2
    distance = ((xx - x) / a) ** 2 + ((yy - y) / b) ** 2

    # Set the pixels within the ellipse to 1, and the pixels outside the ellipse to 0
    ellipse = np.zeros(dims)
    ellipse[distance <= 1] = 1

    return ellipse



import os

from optinist.api.dir_path import DIRPATH
from optinist.api.utils.filepath_creater import join_filepath
from optinist.routers.model import FILETYPE
from optinist.wrappers import wrapper_dict


class SmkUtils:
    @classmethod
    def input(cls, details):
        if details["type"] in [FILETYPE.IMAGE]:
            return [join_filepath([DIRPATH.INPUT_DIR, x]) for x in details["input"]]
        elif details["type"] in [FILETYPE.CSV, FILETYPE.BEHAVIOR, FILETYPE.HDF5]:
            return join_filepath([DIRPATH.INPUT_DIR, details["input"]])
        else:
            return [join_filepath([DIRPATH.OUTPUT_DIR, x]) for x in details["input"]]

    @classmethod
    def output(cls, details):
        return join_filepath([DIRPATH.OUTPUT_DIR, details["output"]])

    @classmethod
    def conda(cls, details):
        if details["type"] in [
            FILETYPE.IMAGE,
            FILETYPE.CSV,
            FILETYPE.BEHAVIOR,
            FILETYPE.HDF5,
        ]:
            return None

        wrapper = cls.dict2leaf(wrapper_dict, details["path"].split("/"))

        if "conda_name" in wrapper:
            _filename = wrapper["conda_name"]
            conda_filepath = f"{DIRPATH.CONDAENV_DIR}/envs/{_filename}"
            if os.path.exists(conda_filepath):
                return conda_filepath

        if "conda_yaml" in wrapper:
            conda_yaml = wrapper["conda_yaml"]
            return f"{DIRPATH.CONDAYML_DIR}/{conda_yaml}"

        return None

    @classmethod
    def dict2leaf(cls, root_dict: dict, path_list):
        path = path_list.pop(0)
        if len(path_list) > 0:
            return cls.dict2leaf(root_dict[path], path_list)
        else:
            return root_dict[path]



import numpy as np
import scipy.signal
import scipy.sparse


def matlab_conv2(x, y, mode="same"):
    if mode == "same":
        # https://stackoverflow.com/a/38355889
        return np.rot90(
            scipy.signal.convolve2d(np.rot90(x, 2), np.rot90(y, 2), mode="same"), 2
        )
    if mode == "valid":
        # https://stackoverflow.com/questions/43270274/equivalent-of-matlab-filter2filter-image-valid-in-python
        # Still there is some difference
        return scipy.signal.convolve2d(x, np.rot90(y, 2), mode="valid")


def insert_binary_col_binary_csc(mat, indices, i=None):
    """
    Insert a column into a binary csc sparse matrix as ith column.
    The column to be inserted is a binary column vector.
    The indices of nonzero elements is given as argument indices.
    if i is None, insert the column as the first column of mat.
    Mayber faster than specified i.

    This method is in-place operation.

    Parameters
    ----------
    mat: scipy.sparse.csc_matrix
    i: int
    indices: numpy.array
        The indices of nonzero elements in the column to be inserted.


    Let mat = scipy.sparse.csc_matrix(np.array([
        [1, 0, 1],
        [1, 1, 1],
        [0, 0, 0],
    ]))
    i = 1
    indices = np.array([2]).
    The column to be inserted is
    [
        [0],
        [0],
        [1],
    ]
    insert_binary_col_binary_csc(mat, indices, i) changes mat to be
    [
        [1, 0, 0, 1],
        [1, 0, 1, 1],
        [0, 1, 0, 0],
    ]       ^ Inserted column

    When indices = np.array([0, 2]),
    The colum to be inserted is
    [
        [1],
        [0],
        [1],
    ]
    and insert_binary_col_binary_csc(mat, indices, i) changes mat to be
    [
        [1, 1, 0, 1],
        [1, 0, 1, 1],
        [0, 1, 0, 0],
    ]       ^ Inserted column
    """

    if not isinstance(mat, scipy.sparse.csc_matrix):
        raise ValueError("works only for CSC format -- use .tocsc() first")
    assert i <= mat.shape[1]
    mat.data = np.ones(len(mat.data) + len(indices), mat.dtype)
    if i is None:
        i = 0
    if i == 0:
        mat.indptr = np.concatenate(
            [np.array([0, len(indices)]), mat.indptr[1:] + len(indices)]
        )
        mat.indices = np.concatenate([indices, mat.indices])
    else:
        mat.indptr = np.concatenate(
            [
                mat.indptr[: i + 1],
                np.array([len(indices) + mat.indptr[i]]),
                len(indices) + mat.indptr[i + 1 :],
            ]
        )
        mat.indices = np.concatenate(
            [mat.indices[: mat.indptr[i]], indices, mat.indices[mat.indptr[i] :]]
        )
    mat._shape = (mat._shape[0], mat._shape[1] + 1)


def delete_row_csr(mat, i):
    """https://stackoverflow.com/a/13078768"""
    mat = mat.copy()
    if not isinstance(mat, scipy.sparse.csr_matrix):
        raise ValueError("works only for CSR format -- use .tocsr() first")
    n = mat.indptr[i + 1] - mat.indptr[i]
    if n > 0:
        mat.data[mat.indptr[i] : -n] = mat.data[mat.indptr[i + 1] :]
        mat.data = mat.data[:-n]
        mat.indices[mat.indptr[i] : -n] = mat.indices[mat.indptr[i + 1] :]
        mat.indices = mat.indices[:-n]
    mat.indptr[i:-1] = mat.indptr[i + 1 :]
    mat.indptr[i:] -= n
    mat.indptr = mat.indptr[:-1]
    mat._shape = (mat._shape[0] - 1, mat._shape[1])
    return mat


def delete_col_csc(mat, i):
    """https://stackoverflow.com/a/13078768"""
    mat = mat.copy()
    if not isinstance(mat, scipy.sparse.csc_matrix):
        raise ValueError("works only for CSC format -- use .tocsc() first")
    n = mat.indptr[i + 1] - mat.indptr[i]
    if n > 0:
        mat.data[mat.indptr[i] : -n] = mat.data[mat.indptr[i + 1] :]
        mat.data = mat.data[:-n]
        mat.indices[mat.indptr[i] : -n] = mat.indices[mat.indptr[i + 1] :]
        mat.indices = mat.indices[:-n]
    mat.indptr[i:-1] = mat.indptr[i + 1 :]
    mat.indptr[i:] -= n
    mat.indptr = mat.indptr[:-1]
    mat._shape = (mat._shape[0], mat._shape[1] - 1)
    return mat


def delete_col_csc_inplace(mat, i):
    """https://stackoverflow.com/a/13078768"""
    if not isinstance(mat, scipy.sparse.csc_matrix):
        raise ValueError("works only for CSC format -- use .tocsc() first")
    n = mat.indptr[i + 1] - mat.indptr[i]
    if n > 0:
        mat.data[mat.indptr[i] : -n] = mat.data[mat.indptr[i + 1] :]
        mat.data = mat.data[:-n]
        mat.indices[mat.indptr[i] : -n] = mat.indices[mat.indptr[i + 1] :]
        mat.indices = mat.indices[:-n]
    mat.indptr[i:-1] = mat.indptr[i + 1 :]
    mat.indptr[i:] -= n
    mat.indptr = mat.indptr[:-1]
    mat._shape = (mat._shape[0], mat._shape[1] - 1)
    return mat


def count_nonzero_values_in_col_csc(mat, i):
    """
    count nonzero values of mat's ith column
    equivalent to np.sum(mat[:, i]>0),
    mat[:, i].count_nonzero()
    and np.count_nonzero(mat[:, i].toarray())
    """
    if not isinstance(mat, scipy.sparse.csc_matrix):
        raise ValueError("works only for CSC format -- use .tocsc() first")
    return mat.indptr[i + 1] - mat.indptr[i]


def intersect_unique_sorted_1d(arr1, arr2):
    """
    Assuming each array is sorted and unique,
    this method is faster than numpy.intersect1d by 1.5 times
    """

    def new_merge(a, b):
        """https://stackoverflow.com/a/54131815"""
        if len(a) < len(b):
            b, a = a, b
        c = np.empty(len(a) + len(b), dtype=a.dtype)
        b_indices = np.arange(len(b)) + np.searchsorted(a, b)
        a_indices = np.ones(len(c), dtype=bool)
        a_indices[b_indices] = False
        c[b_indices] = b
        c[a_indices] = a
        return c

    aux = new_merge(arr1, arr2)
    mask = aux[1:] == aux[:-1]
    int1d = aux[:-1][mask]
    return int1d


def array_to_lil_row(arr):
    """convert 1-dim array into a scipy.sparse.lil style data and row.

    Assume a large matrix's rows are given sequentially.
    This method converts each row into lil style.
    Stacking the results of this method,
    lil style sparse matrix of the large matrix is achieved.


    Parameters
    ----------
    arr: 1 dim numpy.array

    Returns
    -------
    datum: list of lists
    row: list of lists

    Examples
    --------
    >>> arr = np.array([1, 2, 3])
    >>> array_row_to_lil_row(arr)
    list([1, 2, 3]), list([0, 1, 2])

    >>> arr = np.array([0, 0, 1])
    >>> array_row_to_lil_row(arr)
    list([1]), list([2])

    >>> arr = np.array([0, 0, 0])
    >>> array_row_to_lil_row(arr)
    list([]), list([])

    # example of array to sparse.lil_matrix conversion.
    >>> arr = np.array([
    ...     [1, 2, 3],
    ...     [0, 0, 1],
    ...     [0, 0, 0],
    ...     [7, 8, 0],
    ... ])
    >>> rows, data = [], []
    >>> for row_ in arr:
    ...     datum, row = array_to_lil_row(row_)
    ...     rows.append(row)
    ...     data.append(datum)
    ... n_rows, n_cols = 4, 3
    ... lil_arr = scipy.sparse.lil_matrix((n_rows, n_cols), dtype=data[0][0].dtype)
    ... lil_arr.data = np.array(data, dtype='object')
    ... lil_arr.rows = np.array(rows, dtype='object')
    >>> lil_arr.toarray()
    array([[1, 2, 3],
           [0, 0, 1],
           [0, 0, 0],
           [7, 8, 0]])
    """

    row = list(np.where(arr != 0)[0])
    datum = list(arr[arr != 0])
    return datum, row



from optinist.api.dataclass.dataclass import CaimanCnmfData, FluoData, RoiData
from optinist.api.edit_ROI.wrappers.caiman_edit_roi.utils import set_nwbfile


def execute_merge_roi(node_dirpath: str, ids: list):
    import numpy as np

    # load data
    cnmf_data = np.load(f"{node_dirpath}/caiman_cnmf.npy", allow_pickle=True).item()
    is_cell = cnmf_data.get("is_cell")
    merge_roi = cnmf_data.get("merge_roi", [])
    im = cnmf_data.get("im")
    fluorescence = cnmf_data.get("fluorescence")

    # get merging ROI
    merging_ROIs = []
    [merging_ROIs.append(im[id, :, :]) for id in ids]

    # get merged ROI
    merged_ROI = np.maximum.reduce(merging_ROIs)
    is_cell[ids] = False
    is_cell = np.append(is_cell, True)
    im = np.concatenate((im, merged_ROI[np.newaxis, :, :]), axis=0)

    # get merged F
    merged_f = np.mean((fluorescence[ids, :]), axis=0)
    fluorescence = np.vstack([fluorescence, merged_f])

    cell_roi = np.zeros(im.shape)
    num_rois = im.shape[0]
    for i in range(num_rois):
        cell_roi[i, :, :] = np.where(im[i, :, :] != 0, i + 1, np.nan)
    merge_roi.append(float(num_rois))
    merge_roi += [(id + 1) for id in ids]
    merge_roi.append((-1.0))

    cnmf_data["im"] = im
    cnmf_data["is_cell"] = is_cell
    cnmf_data["fluorescence"] = fluorescence
    cnmf_data["merge_roi"] = merge_roi

    info = {
        "fluorescence": FluoData(fluorescence, file_name="fluorescence"),
        "cell_roi": RoiData(
            np.nanmax(cell_roi[is_cell], axis=0),
            output_dir=node_dirpath,
            file_name="cell_roi",
        ),
        "cnmf_data": CaimanCnmfData(cnmf_data),
        "nwbfile": set_nwbfile(cnmf_data),
    }

    return info



from optinist.api.config.config_reader import ConfigReader
from optinist.api.dir_path import DIRPATH
from optinist.api.utils.filepath_creater import join_filepath


def get_typecheck_params(message_params, name):
    default_params = ConfigReader.read(
        join_filepath([DIRPATH.CONFIG_DIR, f"{name}.yaml"])
    )
    if message_params != {} and message_params is not None:
        return check_types(nest2dict(message_params), default_params)
    return default_params


def check_types(params, default_params):
    for key in params.keys():
        if isinstance(params[key], dict):
            params[key] = check_types(params[key], default_params[key])
        else:
            if not isinstance(type(params[key]), type(default_params[key])):
                data_type = type(default_params[key])
                p = params[key]
                if isinstance(data_type, str):
                    params[key] = str(p)
                elif isinstance(data_type, float):
                    params[key] = float(p)
                elif isinstance(data_type, int):
                    params[key] = int(p)

    return params


def nest2dict(value):
    nwb_dict = {}
    for _k, _v in value.items():
        if _v["type"] == "child":
            nwb_dict[_k] = _v["value"]
        elif _v["type"] == "parent":
            nwb_dict[_k] = nest2dict(_v["children"])

    return nwb_dict



import os

import numpy as np

from optinist.api.dataclass.dataclass import RoiData, Suite2pData
from optinist.api.edit_ROI.wrappers.suite2p_edit_roi.utils import set_nwbfile


def excute_delete_roi(node_dirpath, ids):
    ops = np.load(os.path.join(node_dirpath, "suite2p.npy"), allow_pickle=True).item()
    iscell = ops.get("iscell")
    delete_roi = ops.get("delete_roi", [])
    im = ops.get("im")

    delete_roi = ops.get("delete_roi") if ops.get("delete_roi") else []
    for id in ids:
        iscell[id] = False
        delete_roi.append(id + 1)

    ops["iscell"] = iscell
    ops["delete_roi"] = delete_roi

    info = {
        "ops": Suite2pData(ops),
        "non_cell_roi": RoiData(
            np.nanmax(im[~iscell], axis=0),
            output_dir=node_dirpath,
            file_name="noncell_roi",
        ),
        "cell_roi": RoiData(
            np.nanmax(im[iscell], axis=0), output_dir=node_dirpath, file_name="cell_roi"
        ),
        "nwbfile": set_nwbfile(ops),
    }

    return info



import os
import shutil
from glob import glob
from typing import List

from fastapi import APIRouter, File, UploadFile

from optinist.api.dir_path import DIRPATH
from optinist.api.utils.filepath_creater import create_directory, join_filepath
from optinist.routers.const import ACCEPT_CSV_EXT, ACCEPT_HDF5_EXT, ACCEPT_TIFF_EXT
from optinist.routers.model import FILETYPE, FilePath, TreeNode

router = APIRouter()


class DirTreeGetter:
    @classmethod
    def get_tree(cls, file_types: List[str], dirname: str = None) -> List[TreeNode]:
        nodes: List[TreeNode] = []

        if dirname is None:
            absolute_dirpath = DIRPATH.INPUT_DIR
        else:
            absolute_dirpath = join_filepath([DIRPATH.INPUT_DIR, dirname])

        for node_name in os.listdir(absolute_dirpath):
            if dirname is None:
                relative_path = node_name
            else:
                relative_path = join_filepath([dirname, node_name])

            search_dirpath = join_filepath([absolute_dirpath, node_name])

            if os.path.isfile(search_dirpath) and node_name.endswith(tuple(file_types)):
                nodes.append(
                    TreeNode(
                        path=relative_path,
                        name=node_name,
                        isdir=False,
                        nodes=[],
                    )
                )
            elif (
                os.path.isdir(search_dirpath)
                and len(cls.accept_files(search_dirpath, file_types)) > 0
            ):
                nodes.append(
                    TreeNode(
                        path=node_name,
                        name=node_name,
                        isdir=True,
                        nodes=cls.get_tree(file_types, relative_path),
                    )
                )

        return nodes

    @classmethod
    def accept_files(cls, path: str, file_types: List[str]):
        files_list = []
        for file_type in file_types:
            files_list.extend(
                glob(join_filepath([path, "**", f"*{file_type}"]), recursive=True)
            )

        return files_list


@router.get("/files", response_model=List[TreeNode], tags=["files"])
async def get_files(file_type: str = None):
    if file_type == FILETYPE.IMAGE:
        return DirTreeGetter.get_tree(ACCEPT_TIFF_EXT)
    elif file_type == FILETYPE.CSV:
        return DirTreeGetter.get_tree(ACCEPT_CSV_EXT)
    elif file_type == FILETYPE.HDF5:
        return DirTreeGetter.get_tree(ACCEPT_HDF5_EXT)


@router.post("/files/upload/{filename}", response_model=FilePath, tags=["files"])
async def create_file(filename: str, file: UploadFile = File(...)):
    create_directory(DIRPATH.INPUT_DIR)

    filepath = join_filepath([DIRPATH.INPUT_DIR, filename])

    with open(filepath, "wb") as f:
        shutil.copyfileobj(file.file, f)

    return {"file_path": filename}



import os

import numpy as np
import pandas as pd
import tifffile

from optinist.api.utils.filepath_creater import create_directory, join_filepath


class JsonWriter:
    @classmethod
    def write(cls, filepath, data):
        pd.DataFrame(data).to_json(filepath, indent=4)

    @classmethod
    def write_as_split(cls, filepath, data):
        pd.DataFrame(data).to_json(filepath, indent=4, orient="split")


def save_tiff2json(tiff_filepath, save_dirpath, start_index=None, end_index=None):
    # Tiff画像を読み込む
    tiffs = []
    image = tifffile.imread(tiff_filepath)
    if image.ndim == 2:
        image = image[np.newaxis, :, :]

    for i, page in enumerate(image):
        if i < start_index - 1:
            continue

        if i >= end_index:
            break

        tiffs.append(page.tolist())

    filename, _ = os.path.splitext(os.path.basename(tiff_filepath))
    create_directory(save_dirpath)

    JsonWriter.write_as_split(
        join_filepath(
            [save_dirpath, f"{filename}_{str(start_index)}_{str(end_index)}.json"]
        ),
        pd.DataFrame(tiffs),
    )



import numpy as np

from optinist.api.dataclass.dataclass import (
    BehaviorData,
    FluoData,
    HeatMapData,
    IscellData,
    TimeSeriesData,
)
from optinist.api.nwb.nwb import NWBDATASET


def calc_trigger(behavior_data, trigger_type, trigger_threshold):
    flg = np.array(behavior_data > trigger_threshold, dtype=int)
    if trigger_type == "up":
        trigger_idx = np.where(np.ediff1d(flg) == 1)[0]
    elif trigger_type == "down":
        trigger_idx = np.where(np.ediff1d(flg) == -1)[0]
    elif trigger_type == "cross":
        trigger_idx = np.where(np.ediff1d(flg) != 0)[0]
    else:
        trigger_idx = np.where(np.ediff1d(flg) == 0)[0]

    return trigger_idx


def calc_trigger_average(neural_data, trigger_idx, start_time, end_time):
    num_frame = neural_data.shape[0]

    ind = np.array(range(start_time, end_time), dtype=int)

    event_trigger_data = []
    for trigger in trigger_idx:
        target_idx = ind + trigger

        if np.min(target_idx) >= 0 and np.max(target_idx) < num_frame:
            event_trigger_data.append(neural_data[target_idx])

    # (num_event, cell_number, event_time_lambda)
    event_trigger_data = np.array(event_trigger_data)

    return event_trigger_data


def ETA(
    neural_data: FluoData,
    behaviors_data: BehaviorData,
    output_dir: str,
    iscell: IscellData = None,
    params: dict = None,
) -> dict(mean=TimeSeriesData):
    neural_data = neural_data.data
    behaviors_data = behaviors_data.data

    if params["transpose_x"]:
        X = neural_data.transpose()
    else:
        X = neural_data

    if params["transpose_y"]:
        Y = behaviors_data.transpose()
    else:
        Y = behaviors_data

    assert (
        X.shape[0] == Y.shape[0]
    ), f"""
        neural_data and behaviors_data is not same dimension,
        neural.shape{X.shape}, behavior.shape{Y.shape}"""

    if iscell is not None:
        iscell = iscell.data
        ind = np.where(iscell > 0)[0]
        cell_numbers = ind
        X = X[:, ind]

    Y = Y[:, params["target_index"]]

    # calculate Triggers
    trigger_idx = calc_trigger(Y, params["trigger_type"], params["trigger_threshold"])

    # calculate Triggered average
    event_trigger_data = calc_trigger_average(
        X, trigger_idx, params["start_time"], params["end_time"]
    )

    # (cell_number, event_time_lambda)
    if len(event_trigger_data) > 0:
        mean = np.mean(event_trigger_data, axis=0)
        sem = np.std(event_trigger_data, axis=0) / np.sqrt(len(event_trigger_data))
        mean = mean.transpose()
        sem = sem.transpose()
    else:
        assert False, "Output data size is 0"

    # NWB追加
    nwbfile = {}
    nwbfile[NWBDATASET.POSTPROCESS] = {
        "mean": mean,
        "sem": sem,
        "num_sample": [len(mean)],
    }

    min_value = np.min(mean, axis=1, keepdims=True)
    max_value = np.max(mean, axis=1, keepdims=True)
    norm_mean = (mean - min_value) / (max_value - min_value)

    info = {}
    info["mean"] = TimeSeriesData(
        mean,
        std=sem,
        index=list(np.arange(params["start_time"], params["end_time"])),
        cell_numbers=cell_numbers if iscell is not None else None,
        file_name="mean",
    )
    info["mean_heatmap"] = HeatMapData(
        norm_mean,
        columns=list(np.arange(params["start_time"], params["end_time"])),
        file_name="mean_heatmap",
    )
    info["nwbfile"] = nwbfile

    return info



from optinist.api.nwb.nwb import NWBDATASET


def set_nwbfile(cnmf_data):
    im = cnmf_data.get("im")
    is_cell = cnmf_data.get("is_cell")
    fluorescence = cnmf_data.get("fluorescence")

    # NWBの追加
    nwbfile = {}
    # NWBにROIを追加
    roi_list = []
    n_cells = im.shape[0]
    for i in range(n_cells):
        kargs = {}
        kargs["image_mask"] = im[i, :]
        roi_list.append(kargs)

    nwbfile[NWBDATASET.ROI] = {"roi_list": roi_list}

    # iscellを追加
    nwbfile[NWBDATASET.COLUMN] = {
        "roi_column": {
            "name": "iscell",
            "discription": "two columns - iscell & probcell",
            "data": is_cell,
        }
    }

    # Fluorescence

    nwbfile[NWBDATASET.FLUORESCENCE] = {
        "RoiResponseSeries": {
            "table_name": "ROIs",
            "region": list(range(n_cells)),
            "name": "RoiResponseSeries",
            "data": fluorescence.T,
            "unit": "lumens",
        },
    }

    # NWB追加
    nwbfile[NWBDATASET.POSTPROCESS] = {
        "add_roi": cnmf_data.get("add_roi", []),
        "delete_roi": cnmf_data.get("delete_roi", []),
        "merge_roi": cnmf_data.get("merge_roi", []),
    }

    return nwbfile



# from optinist.wrappers.optinist_wrapper.basic_neural_analysis.cell_grouping import (
#     cell_grouping,
# )
from optinist.wrappers.optinist_wrapper.basic_neural_analysis.eta import ETA

basic_neural_analysis_wrapper_dict = {
    "eta": {"function": ETA},
    # 'cell_grouping': {
    #     'function': cell_grouping
    # }
}
